import errno

from leapp.libraries.actor import tcpwrappersconfigread
from leapp.libraries.common.testutils import make_IOError


class MockFileReader(object):
    def __init__(self):
        self.files = {}
        self.files_read = {}
        self.read_called = 0

    def _increment_read_counters(self, path):
        self.read_called += 1
        self.files_read.setdefault(path, 0)
        self.files_read[path] += 1

    def read(self, path):
        self._increment_read_counters(path)
        try:
            return self.files[path]
        except KeyError:
            raise make_IOError(errno.ENOENT)


def test_get_daemon_list_in_line_simple():
    line = 'vsftpd : 192.168.2.*'
    daemon_list = tcpwrappersconfigread._get_daemon_list_in_line(line)
    assert daemon_list.value == ['vsftpd']


def test_get_daemon_list_in_line_multiple():
    line = 'vsftpd, sendmail : 192.168.2.*'
    daemon_list = tcpwrappersconfigread._get_daemon_list_in_line(line)
    assert daemon_list.value == ['vsftpd', 'sendmail']

    line = 'ALL EXCEPT sendmail : 192.168.2.*'
    daemon_list = tcpwrappersconfigread._get_daemon_list_in_line(line)
    assert daemon_list.value == ['ALL', 'EXCEPT', 'sendmail']

    # different separators
    line = 'vsftpd,sendmail : 192.168.2.*'
    daemon_list = tcpwrappersconfigread._get_daemon_list_in_line(line)
    assert daemon_list.value == ['vsftpd', 'sendmail']

    line = 'vsftpd\tsendmail : 192.168.2.*'
    daemon_list = tcpwrappersconfigread._get_daemon_list_in_line(line)
    assert daemon_list.value == ['vsftpd', 'sendmail']

    line = 'vsftpd, \t sendmail : 192.168.2.*'
    daemon_list = tcpwrappersconfigread._get_daemon_list_in_line(line)
    assert daemon_list.value == ['vsftpd', 'sendmail']


def test_get_daemon_list_in_line_malformed():
    line = 'foo'
    daemon_list = tcpwrappersconfigread._get_daemon_list_in_line(line)
    # tcp_wrappers actually ignores lines like this, but there's no harm in being
    # over-sensitive here.
    assert daemon_list.value == ['foo']


def test_get_lines_empty():
    content = ''
    lines = tcpwrappersconfigread._get_lines(content)
    assert lines == ['']


def test_get_lines_simple():
    content = 'vsftpd : 192.168.2.*\n' \
              'ALL : 192.168.1.*\n'
    lines = tcpwrappersconfigread._get_lines(content)
    assert lines == content.split('\n')


def test_get_lines_continued_line():
    content = 'vsftpd : 192.168\\\n.2.*'
    lines = tcpwrappersconfigread._get_lines(content)
    expected = ['vsftpd : 192.168.2.*']
    assert lines == expected


def test_get_lines_backslash_followed_by_whitespace():
    content = 'foo \\ \nthis is not a continuation line'
    lines = tcpwrappersconfigread._get_lines(content)
    expected = ['foo \\ ', 'this is not a continuation line']
    assert lines == expected


def test_get_lines_continued_comment():
    content = '# foo \\\n' \
              'this is still a comment'
    lines = tcpwrappersconfigread._get_lines(content)
    expected = ['# foo this is still a comment']
    assert lines == expected


def test_is_comment():
    assert tcpwrappersconfigread._is_comment('') is True
    assert tcpwrappersconfigread._is_comment('  ') is True
    assert tcpwrappersconfigread._is_comment('# foo') is True
    assert tcpwrappersconfigread._is_comment('#') is True
    assert tcpwrappersconfigread._is_comment(' # foo') is False
    assert tcpwrappersconfigread._is_comment('foo') is False
    assert tcpwrappersconfigread._is_comment(' foo') is False


def test_get_daemon_lists_in_file():
    path = '/etc/hosts.allow'
    reader = MockFileReader()
    reader.files[path] = 'vsftpd : 192.168.2.*\n' \
                         'ALL : 192.168.1.*\n'

    daemon_lists = tcpwrappersconfigread._get_daemon_lists_in_file(path, read_func=reader.read)

    num_lines = 2
    assert len(daemon_lists) == num_lines
    assert daemon_lists[0].value == ['vsftpd']
    assert daemon_lists[1].value == ['ALL']


def test_get_daemon_lists_in_file_nonexistent():
    reader = MockFileReader()
    daemon_lists = tcpwrappersconfigread._get_daemon_lists_in_file('/etc/hosts.allow', read_func=reader.read)
    assert not daemon_lists


def test_get_daemon_lists():
    reader = MockFileReader()
    reader.files['/etc/hosts.allow'] = 'vsftpd : 192.168.2.*\n' \
                                       'ALL : 192.168.1.*\n'
    reader.files['/etc/hosts.deny'] = 'sendmail : 192.168.2.*\n'

    daemon_lists = tcpwrappersconfigread._get_daemon_lists(read_func=reader.read)

    num_lines = 3
    assert len(daemon_lists) == num_lines
    assert daemon_lists[0].value == ['vsftpd']
    assert daemon_lists[1].value == ['ALL']
    assert daemon_lists[2].value == ['sendmail']


def test_get_daemon_lists_nonexistent_config():
    reader = MockFileReader()
    daemon_lists = tcpwrappersconfigread._get_daemon_lists(read_func=reader.read)
    assert not daemon_lists


def test_get_tcp_wrappers_facts():
    reader = MockFileReader()
    reader.files['/etc/hosts.allow'] = 'vsftpd : 192.168.2.*\n' \
                                       'ALL : 192.168.1.*\n'
    reader.files['/etc/hosts.deny'] = 'sendmail : 192.168.2.*\n'

    facts = tcpwrappersconfigread.get_tcp_wrappers_facts(read_func=reader.read)

    num_lines = 3
    assert len(facts.daemon_lists) == num_lines
    assert facts.daemon_lists[0].value == ['vsftpd']
    assert facts.daemon_lists[1].value == ['ALL']
    assert facts.daemon_lists[2].value == ['sendmail']


def test_get_tcp_wrappers_facts_nonexistent_config():
    reader = MockFileReader()
    facts = tcpwrappersconfigread.get_tcp_wrappers_facts(read_func=reader.read)
    assert not facts.daemon_lists



from leapp.libraries.common.rpms import has_package
from leapp.libraries.stdlib import api
from leapp.models import InstalledRPM


def get_kde_apps_info():
    installed = list()
    base_kde_apps = ("kde-baseapps",
                     "okular",
                     "ark",
                     "kdepim",
                     "konsole",
                     "gwenview",
                     "kdenetwork",
                     "kate",
                     "kwrite")

    api.current_logger().info("  Detecting installed KDE apps  ")
    api.current_logger().info("================================")
    for app in [application for application in base_kde_apps if has_package(InstalledRPM, application)]:
        api.current_logger().info("Application {0} is installed.".format(app))
        installed.append(app)
    api.current_logger().info("----------------------------------")

    return installed



from leapp.models import SpamassassinFacts
from leapp.reporting import Report
from leapp.snactor.fixture import current_actor_context


def test_actor_basic(current_actor_context):
    facts = SpamassassinFacts(service_overriden=False)

    current_actor_context.feed(facts)
    current_actor_context.run()
    reports = current_actor_context.consume(Report)

    assert len(reports) == 4
    report = reports[0]
    assert '--ssl' in report.report['summary']
    assert 'spamc' in report.report['summary']
    report = reports[1]
    assert '--ssl-version' in report.report['summary']
    assert 'spamd' in report.report['summary']
    report = reports[2]
    assert 'spamassassin.service' in report.report['summary']
    report = reports[3]
    assert 'sa-update no longer supports SHA1' in report.report['summary']


def test_actor_no_message(current_actor_context):
    current_actor_context.run()
    assert not current_actor_context.consume(Report)



from leapp import reporting
from leapp.libraries.common.spamassassinutils import (
    SPAMASSASSIN_SERVICE_OVERRIDE,
    SPAMC_CONFIG_FILE,
    SYSCONFIG_SPAMASSASSIN
)


def _check_spamc_config(facts, report_func):
    title = ('spamc no longer allows specifying the TLS version and no longer '
             'supports SSLv3')
    summary_generic = ('spamc no longer allows using the "--ssl" option with an '
                       'argument specifying the TLS version - the option can only '
                       'be used without an argument. Also, spamc no longer supports '
                       'SSLv3.')
    if facts.spamc_ssl_argument:
        summary_detail = ('The spamc configuration file uses "--ssl %s", it will '
                          'be updated during the upgrade.'
                          % facts.spamc_ssl_argument)
        summary = reporting.Summary('%s %s' % (summary_generic, summary_detail))
        resource = reporting.RelatedResource('file', SPAMC_CONFIG_FILE)
    else:
        summary = reporting.Summary(summary_generic)
        resource = None
    severity = (reporting.Severity.HIGH if facts.spamc_ssl_argument == 'sslv3'
                else reporting.Severity.MEDIUM)
    hint = 'Please update your scripts and configuration, if there are any.'

    args = [
        reporting.Title(title),
        summary,
        reporting.Groups([reporting.Groups.ENCRYPTION]),
        reporting.Severity(severity),
        reporting.Remediation(hint=hint),
    ]
    if resource:
        args.append(resource)
    report_func(args)


def _check_spamd_config_ssl(facts, report_func):
    title = ('spamd no longer allows specifying the TLS version and no longer '
             'supports SSLv3')
    summary_generic = ('spamd no longer accepts the --ssl-version option and '
                       'no longer supports SSLv3.')
    if facts.spamd_ssl_version:
        summary_detail = ('The spamd sysconfig file uses "--ssl-version %s", '
                          'it will be updated during the upgrade.'
                          % facts.spamd_ssl_version)
        summary = reporting.Summary('%s %s' % (summary_generic, summary_detail))
        resource = reporting.RelatedResource('file', SYSCONFIG_SPAMASSASSIN)
    else:
        summary = reporting.Summary(summary_generic)
        resource = None
    severity = (reporting.Severity.HIGH if facts.spamd_ssl_version == 'sslv3'
                else reporting.Severity.MEDIUM)
    hint = 'Please update your scripts and configuration, if there are any.'

    args = [
        reporting.Title(title),
        summary,
        reporting.Groups([reporting.Groups.ENCRYPTION, reporting.Groups.SERVICES]),
        reporting.Severity(severity),
        reporting.Remediation(hint=hint)
    ]
    if resource:
        args.append(resource)
    report_func(args)


def _check_spamd_config_service_type(facts, report_func):
    title = 'The type of the spamassassin systemd service has changed'
    summary_generic = 'The type of spamassassin.service has been changed from "forking" to "simple".'
    if facts.service_overriden:
        summary_detail = 'However, the service appears to be overridden; no migration action will occur.'
        resource = reporting.RelatedResource('file', SPAMASSASSIN_SERVICE_OVERRIDE)
    else:
        summary_detail = 'The spamassassin sysconfig file will be updated.'
        resource = reporting.RelatedResource('file', SYSCONFIG_SPAMASSASSIN)
    report_func([
        reporting.Title(title),
        reporting.Summary('%s %s' % (summary_generic, summary_detail)),
        reporting.Groups([reporting.Groups.SERVICES]),
        reporting.Severity(reporting.Severity.MEDIUM),
        resource
    ])


def _report_sa_update_change(report_func):
    summary = ('sa-update no longer supports SHA1 validation of filtering rules, '
               'SHA256/SHA512 validation is done instead. This may affect you if '
               'you are using an alternative update channel (sa-update used with '
               'option --channel or --channelfile), or if you install filtering '
               'rule updates directly from files (sa-update --install).')
    report_func([reporting.Title('sa-update no longer supports SHA1 validation'),
                 reporting.Summary(summary),
                 reporting.Severity(reporting.Severity.LOW)])


def produce_reports(facts):
    """
    Checks spamassassin configuration and produces reports.
    """
    _check_spamc_config(facts, reporting.create_report)
    _check_spamd_config_ssl(facts, reporting.create_report)
    _check_spamd_config_service_type(facts, reporting.create_report)
    _report_sa_update_change(reporting.create_report)



from leapp import reporting
from leapp.libraries.common.config.architecture import ARCH_X86_64, matches_architecture
from leapp.libraries.stdlib import api
from leapp.models import CPUInfo

X86_64_BASELINE_FLAGS = ['cmov', 'cx8', 'fpu', 'fxsr', 'mmx', 'syscall', 'sse', 'sse2']
X86_64_V2_FLAGS = ['cx16', 'lahf_lm', 'popcnt', 'pni', 'sse4_1', 'sse4_2', 'ssse3']


def _inhibit_upgrade(missing_flags):
    title = 'Current x86-64 microarchitecture is unsupported in RHEL9'
    summary = ('RHEL9 has a higher CPU requirement than older versions, it now requires a CPU '
               'compatible with x86-64-v2 instruction set or higher.\n\n'
               'Missings flags detected are: {}\n'.format(', '.join(missing_flags)))

    reporting.create_report([
        reporting.Title(title),
        reporting.Summary(summary),
        reporting.ExternalLink(title='Building Red Hat Enterprise Linux 9 for the x86-64-v2 microarchitecture level',
                               url='https://red.ht/rhel-9-intel-microarchitectures'),
        reporting.Severity(reporting.Severity.HIGH),
        reporting.Groups([reporting.Groups.INHIBITOR]),
        reporting.Groups([reporting.Groups.SANITY]),
        reporting.Remediation(hint=('If case of using virtualization, virtualization platforms often allow '
                                    'configuring a minimum denominator CPU model for compatibility when migrating '
                                    'between different CPU models. Ensure that minimum requirements are not below '
                                    'that of RHEL9\n')),
    ])


def process():
    """
    Check whether the processor matches the required microarchitecture.
    """

    if not matches_architecture(ARCH_X86_64):
        api.current_logger().info('Architecture not x86-64. Skipping microarchitecture test.')
        return

    cpuinfo = next(api.consume(CPUInfo))

    required_flags = X86_64_BASELINE_FLAGS + X86_64_V2_FLAGS
    missing_flags = [flag for flag in required_flags if flag not in cpuinfo.flags]
    api.current_logger().debug('Required flags missing: %s', missing_flags)
    if missing_flags:
        _inhibit_upgrade(missing_flags)



import re
from collections import defaultdict, namedtuple

from leapp import reporting
from leapp.libraries.common.config import get_consumed_data_stream_id
from leapp.libraries.common.fetch import ASSET_PROVIDED_DATA_STREAMS_FIELD
from leapp.libraries.common.rpms import get_leapp_packages, LeappComponents
from leapp.libraries.stdlib import api
from leapp.models import ConsumedDataAsset


def _get_hint():
    hint = (
        'All official assets (data files) are part of the installed rpms these days.'
        ' This issue is usually encountered when the data files are incorrectly'
        ' customized, replaced, or removed. '
        ' In case you want to recover the original files, remove them (if they still exist)'
        ' and reinstall the following rpms: {rpms}.\n'
        'The listed assets (data files) are usually inside the /etc/leapp/files/'
        ' directory.'
        .format(
            rpms=', '.join(get_leapp_packages(component=LeappComponents.REPOSITORY))
        )
    )
    return hint


def compose_summary_for_incompatible_assets(assets, incompatibility_reason):
    if not assets:
        return []

    summary_lines = ['The following assets are {reason}'.format(reason=incompatibility_reason)]
    for asset in assets:
        if asset.provided_data_streams is None:  # Assets with missing streams are placed only in .outdated bucket
            details = (' - The asset {what_asset} is missing information about provided data streams '
                       'in its metadata header')
            details = details.format(what_asset=asset.filename)
        else:
            article, multiple_suffix = ('the ', '') if len(asset.provided_data_streams) == 1 else ('', 's')
            details = ' - The asset {what_asset} provides {article}data stream{mult_suffix} {provided_streams}'
            details = details.format(what_asset=asset.filename,
                                     provided_streams=', '.join(asset.provided_data_streams),
                                     article=article, mult_suffix=multiple_suffix)
        summary_lines.append(details)
    return summary_lines


def make_report_entries_with_unique_urls(docs_url_to_title_map):
    report_urls = []
    # Add every unique asset URL into the report
    urls_with_multiple_titles = []
    for url, titles in docs_url_to_title_map.items():
        if len(titles) > 1:
            urls_with_multiple_titles.append(url)

        report_entry = reporting.ExternalLink(title=titles[0], url=url)
        report_urls.append(report_entry)

    if urls_with_multiple_titles:
        msg = 'Docs URLs {urls} are used with inconsistent URL titles, picking one.'
        api.current_logger().warning(msg.format(urls=', '.join(urls_with_multiple_titles)))

    return report_urls


def report_incompatible_assets(assets):
    if not any((assets.outdated, assets.too_new, assets.unknown)):
        return

    title = 'Incompatible Leapp data assets are present'

    docs_url_to_title_map = defaultdict(list)
    required_data_stream = get_consumed_data_stream_id()
    summary_prelude = ('The currently installed Leapp consumes data stream {consumed_data_stream}, but the '
                       'following assets provide different streams:')
    summary_lines = [summary_prelude.format(consumed_data_stream=required_data_stream)]

    assets_with_shared_summary_entry = [
        ('outdated', assets.outdated),
        ('intended for a newer leapp', assets.too_new),
        ('has an incorrect version', assets.unknown)
    ]

    doc_url_to_title = defaultdict(list)  # To make sure we do not spam the user with the same URLs
    for reason, incompatible_assets in assets_with_shared_summary_entry:
        summary_lines += compose_summary_for_incompatible_assets(incompatible_assets, reason)

        for asset in incompatible_assets:
            if asset.docs_url:
                # Add URLs only when they are specified. docs_url could be empty string
                doc_url_to_title[asset.docs_url].append(asset.docs_title)

    report_parts = [
        reporting.Title(title),
        reporting.Summary('\n'.join(summary_lines)),
        reporting.Severity(reporting.Severity.HIGH),
        reporting.Remediation(hint=_get_hint()),
        reporting.Groups([reporting.Groups.INHIBITOR, reporting.Groups.SANITY]),
    ]

    report_parts += make_report_entries_with_unique_urls(docs_url_to_title_map)
    reporting.create_report(report_parts)


def report_malformed_assets(malformed_assets):
    if not malformed_assets:
        return

    title = 'Detected malformed Leapp data assets'
    summary_lines = ['The following assets are malformed:']

    docs_url_to_title_map = defaultdict(list)
    for asset in malformed_assets:
        if not asset.provided_data_streams:
            details = (' - The asset file {filename} contains no values in its "{provided_data_streams_field}" '
                       'field, or the field does not contain a list')
            details = details.format(filename=asset.filename,
                                     provided_data_streams_field=ASSET_PROVIDED_DATA_STREAMS_FIELD)
        else:
            # The asset is malformed because we failed to convert its major versions to ints
            details = ' - The asset file {filename} contains invalid value in its "{data_streams_field}"'
            details = details.format(filename=asset.filename, data_streams_field=ASSET_PROVIDED_DATA_STREAMS_FIELD)
        summary_lines.append(details)
        if asset.docs_url:
            # Add URLs only when they are specified. docs_url could be empty string
            docs_url_to_title_map[asset.docs_url].append(asset.docs_title)

    report_parts = [
        reporting.Title(title),
        reporting.Summary('\n'.join(summary_lines)),
        reporting.Remediation(hint=_get_hint()),
        reporting.Severity(reporting.Severity.HIGH),
        reporting.Groups([reporting.Groups.INHIBITOR, reporting.Groups.SANITY]),
    ]

    report_parts += make_report_entries_with_unique_urls(docs_url_to_title_map)
    reporting.create_report(report_parts)


def inhibit_if_assets_with_incorrect_version():
    required_data_stream = get_consumed_data_stream_id()
    required_data_stream_major = int(required_data_stream.split('.', 1)[0])

    # The assets are collected according to why are they considered incompatible, so that a single report is created
    # for every class
    IncompatibleAssetsByType = namedtuple('IncompatibleAssets', ('outdated', 'too_new', 'malformed', 'unknown'))
    incompatible_assets = IncompatibleAssetsByType(outdated=[], too_new=[], malformed=[], unknown=[])

    datastream_version_re = re.compile(r'\d+\.\d+$')

    for consumed_asset in api.consume(ConsumedDataAsset):
        if consumed_asset.provided_data_streams is None:  # There is no provided_data_streams field
            # Most likely an old file that predates the introduction of versioning to data assets
            incompatible_assets.outdated.append(consumed_asset)
            continue

        # Ignore minor stream numbers and search only for a stream matching the same major number
        if all((datastream_version_re.match(stream) for stream in consumed_asset.provided_data_streams)):
            provided_major_data_streams = sorted(
                int(stream.split('.', 1)[0]) for stream in consumed_asset.provided_data_streams
            )
        else:
            incompatible_assets.malformed.append(consumed_asset)
            continue

        if required_data_stream_major in provided_major_data_streams:
            continue

        if not provided_major_data_streams:
            # The field contained [], or something that was not a list, but it was corrected to [] to satisfy model
            incompatible_assets.malformed.append(consumed_asset)
            continue

        if required_data_stream_major > max(provided_major_data_streams):
            incompatible_assets.outdated.append(consumed_asset)
        elif required_data_stream_major < min(provided_major_data_streams):
            incompatible_assets.too_new.append(consumed_asset)
        else:
            # Since the `provided_data_vers` is a list of values, it is possible that the asset provide, e.g., 4.0
            # and 6.0,  but the leapp consumes 5.0, thus we need to be careful when to say that an asset is too
            # new/outdated/none.
            incompatible_assets.unknown.append(consumed_asset)

    report_incompatible_assets(incompatible_assets)
    report_malformed_assets(incompatible_assets.malformed)



from leapp.actors import Actor
from leapp.libraries.common.rpms import get_installed_rpms
from leapp.models import InstalledUnsignedRPM, LeftoverPackages, RPM, TransactionCompleted
from leapp.tags import IPUWorkflowTag, RPMUpgradePhaseTag


class CheckLeftoverPackages(Actor):
    """
    Check if there are any RHEL 7 packages present after upgrade.

    Actor produces message containing these packages. Message is empty if there are no el7 package left.
    """

    name = 'check_leftover_packages'
    consumes = (TransactionCompleted, InstalledUnsignedRPM)
    produces = (LeftoverPackages,)
    tags = (RPMUpgradePhaseTag, IPUWorkflowTag)

    def process(self):
        LEAPP_PACKAGES = ['leapp', 'leapp-repository', 'snactor', 'leapp-repository-deps-el8', 'leapp-deps-el8',
                          'python2-leapp']
        installed_rpms = get_installed_rpms()
        if not installed_rpms:
            return

        to_remove = LeftoverPackages()
        unsigned = [pkg.name for pkg in next(self.consume(InstalledUnsignedRPM), InstalledUnsignedRPM()).items]

        for rpm in installed_rpms:
            rpm = rpm.strip()
            if not rpm:
                continue
            name, version, release, epoch, packager, arch, pgpsig = rpm.split('|')

            if 'el7' in release and name not in set(unsigned + LEAPP_PACKAGES):
                to_remove.items.append(RPM(
                    name=name,
                    version=version,
                    epoch=epoch,
                    packager=packager,
                    arch=arch,
                    release=release,
                    pgpsig=pgpsig
                ))

        self.produce(to_remove)



from collections import namedtuple

import pytest

import leapp.models
from leapp.libraries.common import dnfplugin
from leapp.libraries.common.config.version import get_major_version
from leapp.libraries.common.testutils import CurrentActorMocked
from leapp.libraries.stdlib import api
from leapp.models.fields import Boolean
from leapp.topics import Topic


class DATADnfPluginDataTopic(Topic):
    name = 'data_dnf_plugin_data'


fields = leapp.models.fields

TaskData = namedtuple('TaskData', 'expected initdata')

TEST_INSTALL_PACKAGES = TaskData(
    expected=('install1', 'install2'),
    initdata=('install1', 'install2')
)
TEST_REMOVE_PACKAGES = TaskData(
    expected=('remove1', 'remove2'),
    initdata=('remove1', 'remove2'),
)
TEST_UPGRADE_PACKAGES = TaskData(
    expected=('upgrade1', 'upgrade2'),
    initdata=('upgrade1', 'upgrade2'),
)
TEST_ENABLE_MODULES = TaskData(
    expected=('enable1:stream1', 'enable2:stream2'),
    initdata=(
        leapp.models.Module(name='enable1', stream='stream1'),
        leapp.models.Module(name='enable2', stream='stream2'),
    )
)


class DATADnfPluginDataPkgsInfo(leapp.models.Model):
    topic = DATADnfPluginDataTopic
    local_rpms = fields.List(fields.String())
    to_install = fields.List(fields.StringEnum(choices=TEST_INSTALL_PACKAGES.expected))
    to_remove = fields.List(fields.StringEnum(choices=TEST_REMOVE_PACKAGES.expected))
    to_upgrade = fields.List(fields.StringEnum(choices=TEST_UPGRADE_PACKAGES.expected))
    modules_to_enable = fields.List(fields.StringEnum(choices=TEST_ENABLE_MODULES.expected))


TEST_ENABLE_REPOS_CHOICES = ('enabled_repo', 'BASEOS', 'APPSTREAM')


class BooleanEnum(fields.EnumMixin, Boolean):
    pass


class DATADnfPluginDataDnfConf(leapp.models.Model):
    topic = DATADnfPluginDataTopic
    allow_erasing = BooleanEnum(choices=[True])
    best = BooleanEnum(choices=[True])
    debugsolver = fields.Boolean()
    disable_repos = BooleanEnum(choices=[True])
    enable_repos = fields.List(fields.StringEnum(choices=TEST_ENABLE_REPOS_CHOICES))
    gpgcheck = fields.Boolean()
    platform_id = fields.StringEnum(choices=['platform:el8', 'platform:el9'])
    releasever = fields.String()
    installroot = fields.StringEnum(choices=['/installroot'])
    test_flag = fields.Boolean()


class DATADnfPluginDataRHUIAWS(leapp.models.Model):
    topic = DATADnfPluginDataTopic
    on_aws = fields.Boolean()
    region = fields.Nullable(fields.String())


class DATADnfPluginDataRHUI(leapp.models.Model):
    topic = DATADnfPluginDataTopic
    aws = fields.Model(DATADnfPluginDataRHUIAWS)


class DATADnfPluginData(leapp.models.Model):
    topic = DATADnfPluginDataTopic
    pkgs_info = fields.Model(DATADnfPluginDataPkgsInfo)
    dnf_conf = fields.Model(DATADnfPluginDataDnfConf)
    rhui = fields.Model(DATADnfPluginDataRHUI)


# Delete those models from leapp.models to 'unpolute' the module
del leapp.models.DATADnfPluginDataPkgsInfo
del leapp.models.DATADnfPluginDataDnfConf
del leapp.models.DATADnfPluginDataRHUI
del leapp.models.DATADnfPluginDataRHUIAWS
del leapp.models.DATADnfPluginData


_CONFIG_BUILD_TEST_DEFINITION = (
    #   Parameter, Input Data, Expected Fields with data
    ('debug', False, ('dnf_conf', 'debugsolver'), False),
    ('debug', True, ('dnf_conf', 'debugsolver'), True),
    ('target_repoids', TEST_ENABLE_REPOS_CHOICES, ('dnf_conf', 'enable_repos'), list(TEST_ENABLE_REPOS_CHOICES)),
    ('target_repoids', TEST_ENABLE_REPOS_CHOICES[0:1],
     ('dnf_conf', 'enable_repos'), list(TEST_ENABLE_REPOS_CHOICES[0:1])),
    ('target_repoids', TEST_ENABLE_REPOS_CHOICES[1:],
     ('dnf_conf', 'enable_repos'), list(TEST_ENABLE_REPOS_CHOICES[1:])),
    ('target_repoids', TEST_ENABLE_REPOS_CHOICES[2:],
     ('dnf_conf', 'enable_repos'), list(TEST_ENABLE_REPOS_CHOICES[2:])),
    ('test', False, ('dnf_conf', 'test_flag'), False),
    ('test', True, ('dnf_conf', 'test_flag'), True),
)


@pytest.mark.parametrize('used_target_version', ['8.4', '8.5', '9.0', '9.1'])
@pytest.mark.parametrize('parameter,input_value,test_path,expected_value', _CONFIG_BUILD_TEST_DEFINITION)
def test_build_plugin_data_variations(
    monkeypatch,
    used_target_version,
    parameter,
    input_value,
    test_path,
    expected_value,
):
    used_target_major_version = get_major_version(used_target_version)
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(dst_ver=used_target_version))
    inputs = {
        'target_repoids': ['BASEOS', 'APPSTREAM'],
        'debug': True,
        'test': True,
        'on_aws': False,
        'tasks': leapp.models.FilteredRpmTransactionTasks(
            to_install=TEST_INSTALL_PACKAGES.initdata,
            to_remove=TEST_REMOVE_PACKAGES.initdata,
            to_upgrade=TEST_UPGRADE_PACKAGES.initdata,
            modules_to_enable=TEST_ENABLE_MODULES.initdata
            )
    }
    inputs[parameter] = input_value
    created = DATADnfPluginData.create(
        dnfplugin.build_plugin_data(
            **inputs
        )
    )
    assert created.dnf_conf.platform_id == 'platform:el{}'.format(used_target_major_version)
    assert created.dnf_conf.releasever == used_target_version
    value = created
    for path in test_path:
        value = getattr(value, path)
    assert value == expected_value


def test_build_plugin_data(monkeypatch):
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(dst_ver='8.4'))
    # Use leapp to validate format and data
    created = DATADnfPluginData.create(
        dnfplugin.build_plugin_data(
            target_repoids=['BASEOS', 'APPSTREAM'],
            debug=True,
            test=True,
            on_aws=False,
            tasks=leapp.models.FilteredRpmTransactionTasks(
                to_install=TEST_INSTALL_PACKAGES.initdata,
                to_remove=TEST_REMOVE_PACKAGES.initdata,
                to_upgrade=TEST_UPGRADE_PACKAGES.initdata,
                modules_to_enable=TEST_ENABLE_MODULES.initdata
                )
            )
    )
    assert created.dnf_conf.debugsolver is True
    assert created.dnf_conf.test_flag is True
    assert created.rhui.aws.on_aws is False

    with pytest.raises(fields.ModelViolationError):
        DATADnfPluginData.create(
            dnfplugin.build_plugin_data(
                target_repoids=['BASEOS', 'APPSTREAM'],
                debug=True,
                test=True,
                on_aws=False,
                tasks=leapp.models.FilteredRpmTransactionTasks(
                    to_install=TEST_INSTALL_PACKAGES.initdata,
                    to_remove=TEST_REMOVE_PACKAGES.initdata,
                    to_upgrade=TEST_UPGRADE_PACKAGES.initdata,
                    # Enforcing the failure
                    modules_to_enable=(
                        leapp.models.Module(
                            name='broken', stream=None
                        ),
                    ),
                )
            )
        )



import json
import os
import re

from leapp.exceptions import StopActorExecutionError
from leapp.libraries.common.rpms import has_package
from leapp.libraries.stdlib import api, CalledProcessError, run
from leapp.models import InstalledRPM

CEPH_CONF = "/etc/ceph/ceph.conf"
CONTAINER = "ceph-osd"


def select_osd_container(engine):
    try:
        output = run([engine, 'ps'])
    except CalledProcessError as cpe:
        raise StopActorExecutionError(
            'Could not retrieve running containers list',
            details={'details': 'An exception raised during containers listing {}'.format(str(cpe))}
        )
    for line in output['stdout'].splitlines():
        container_name = line.split()[-1]
        if re.match(CONTAINER, container_name):
            return container_name
    return None


def get_ceph_lvm_list():
    base_cmd = ['ceph-volume', 'lvm', 'list', '--format', 'json']
    container_binary = 'podman' if has_package(InstalledRPM, 'podman') else \
        'docker' if has_package(InstalledRPM, 'docker') else ''
    if container_binary == '' and has_package(InstalledRPM, 'ceph-osd'):
        cmd_ceph_lvm_list = base_cmd
    elif container_binary == '':
        return None
    else:
        container_name = select_osd_container(container_binary)
        if container_name is None:
            return None
        cmd_ceph_lvm_list = [container_binary, 'exec', container_name]
        cmd_ceph_lvm_list.extend(base_cmd)
    try:
        output = run(cmd_ceph_lvm_list)
    except CalledProcessError as cpe:
        raise StopActorExecutionError(
            'Could not retrieve the ceph volumes list',
            details={'details': 'An exception raised while retrieving ceph volumes {}'.format(str(cpe))}
        )
    try:
        json_output = json.loads(output['stdout'])
    except ValueError as jve:
        raise StopActorExecutionError(
            'Could not load json file containing ceph volume list',
            details={'details': 'json file wrong format {}'.format(str(jve))}
        )
    return json_output


def encrypted_osds_list():
    result = []
    if os.path.isfile(CEPH_CONF):
        output = get_ceph_lvm_list()
        if output is not None:
            try:
                result = [output[key][0]['lv_uuid'] for key in output if output[key][0]['tags']['ceph.encrypted']]
            except KeyError:
                # TODO: possibly raise a report item with a medium risk factor
                # TODO: possibly create list of problematic osds, extend the cephinfo
                # #     model to include the list and then report it.
                api.current_logger().warning('ceph-osd is installed but no encrypted osd has been found')
    return result



import errno
import re

from leapp.libraries.stdlib import api

_sections = ('defaults', 'blacklist', 'blacklist_exceptions', 'devices',
             'overrides', 'multipaths')

_subsections = {'blacklist': 'device', 'blacklist_exceptions': 'device',
                'devices': 'device', 'multipaths': 'multipath'}


def read_config(path):
    try:
        with open(path, 'r') as f:
            return f.read()
    except IOError as e:
        if e.errno == errno.ENOENT:
            api.current_logger().debug(
                'multipath configuration file {} does not exist.'.format(path)
            )
        else:
            api.current_logger().warning(
                'Failed to read multipath configuration file {}: {}'.
                format(path, e)
            )
        return None


def write_config(path, contents):
    try:
        with open(path, 'w') as f:
            f.write(contents)
    except IOError as e:
        api.current_logger().warning(
            'Failed to write multipath configuration file {}: {}'.
            format(path, e)
        )


class LineData(object):
    TYPE_BLANK = 0
    TYPE_SECTION_START = 1
    TYPE_SECTION_END = 2
    TYPE_OPTION = 3

    def __init__(self, line, section, in_subsection):
        comment_pattern = re.compile('^([^"#!]*)[#!]')
        string_pattern = re.compile('^([^"]*)"([^"]*)')
        utf8_pattern = re.compile('[^\x00-\x7F]+')
        line_pattern = re.compile(r'^([^\s{}]+)\s*(\S*)')
        value = None

        r = comment_pattern.match(line)
        if r:
            line = r.group(1)
        r = string_pattern.match(line)
        if r:
            line = r.group(1)
            value = r.group(2)
        line = utf8_pattern.sub(' ', line)
        line = line.strip()

        if line == '':
            self.type = self.TYPE_BLANK
            return
        if line[0] == '}':
            self.type = self.TYPE_SECTION_END
            return

        r = line_pattern.match(line)
        if r is None:
            raise ValueError
        keyword = r.group(1)
        if r.group(2) != '':
            value = r.group(2)  # even if value was set before

        if section is None:
            if keyword in _sections:
                self.type = self.TYPE_SECTION_START
                self.section = keyword
                return
            raise ValueError

        if not in_subsection and section in _subsections and \
                keyword == _subsections[section]:
            self.type = self.TYPE_SECTION_START
            self.section = keyword
            return

        if value is None:
            raise ValueError
        self.type = self.TYPE_OPTION
        self.option = keyword
        self.value = value

    def is_enabled(self):
        if self.value in ('yes', '1'):
            return True
        if self.value in ('no', '0'):
            return False
        return None



from leapp.libraries.common.rpms import has_package
from leapp.models import InstalledRPM
from leapp.reporting import create_report, Groups, RelatedResource, Remediation, Severity, Summary, Title

# package_name: remedy information
PROBLEM_PACKAGE_MAP = {
    'brlapi.i686': {'bugzilla': None},
    'gnome-online-accounts-devel.i686': {
        'bugzilla': 'https://bugzilla.redhat.com/show_bug.cgi?id=1765627'},
    'geocode-glib-devel.i686': {
        'bugzilla': 'https://bugzilla.redhat.com/show_bug.cgi?id=1765629'}}


def check():
    actual_problems = []
    related_resources = []
    for package, details in PROBLEM_PACKAGE_MAP.items():
        name, arch = package.split('.')
        if has_package(InstalledRPM, name, arch) and has_package(InstalledRPM, name, 'x86_64'):
            actual_problems.append(package)
            # generate RelatedResources for the report
            related_resources.append(RelatedResource('package', package))
            if details['bugzilla']:
                related_resources.append(RelatedResource('bugzilla', details['bugzilla']))

    if actual_problems:
        remediation = ["yum", "remove", "-y"] + actual_problems
        # create a single report entry for all problematic packages
        create_report([
            Title('Some packages have both 32bit and 64bit version installed which are known '
                  'to cause rpm transaction test to fail'),
            Summary('The following packages have both 32bit and 64bit version installed which are known '
                    'to cause rpm transaction test to fail:\n{}'.format(
                        '\n'.join(['- {}'.format(a) for a in actual_problems]))),
            Severity(Severity.HIGH),
            Groups([Groups.INHIBITOR]),
            Remediation(commands=[remediation])] + related_resources)



import pytest

from leapp.exceptions import StopActorExecutionError
from leapp.libraries.actor import opensshsubsystemsftp
from leapp.models import OpenSshConfig, Report


def test_no_config(current_actor_context):
    with pytest.raises(StopActorExecutionError):
        opensshsubsystemsftp.process(iter([]))


@pytest.mark.parametrize('modified,subsystem,expected_report', [
    (False, None, False),  # should not happen
    (False, '/usr/libexec/openssh/sftp-server', False),  # Defaults
    (True, None, True),
    (True, 'internal-sftp', False),
    (True, '/usr/libexec/openssh/sftp-server', False)
])
def test_subsystem(current_actor_context, modified, subsystem, expected_report):
    conf = OpenSshConfig(
        modified=modified,
        permit_root_login=[],
        deprecated_directives=[]
    )
    if subsystem is not None:
        conf.subsystem_sftp = subsystem
    current_actor_context.feed(conf)
    current_actor_context.run()
    if expected_report:
        assert current_actor_context.consume(Report)
    else:
        assert not current_actor_context.consume(Report)



import pytest

from leapp.libraries.actor.vimmigrate import update_vim, vim_configs

packages = [
    {
        'vim-minimal': '/etc/virc',
        'vim-enhanced': '/etc/vimrc'
    },
    {
        'vim-minimal': '/etc/virc',
        'vim-enhanced': ''
    },
    {
        'vim-minimal': '',
        'vim-enhanced': '/etc/vimrc'
    },
    {
        'vim-minimal': '',
        'vim-enhanced': ''
    },
    {
        'vim-minimal': '/etc/virc',
        'ble': ''
    },
    {
        'vim-minimal': '',
        'ble': ''
    },
    {
        'vim-enhanced': '/etc/vimrc',
        'moo': ''
    },
    {
        'vim-enhanced': '',
        'moo': ''
    },
    {
        'you': '',
        'hele': ''
    }
]


class MockLogger(object):
    def __init__(self):
        self.debugmsg = ''
        self.errmsg = ''

    def debug(self, message):
        self.debugmsg += message

    def error(self, message):
        self.errmsg += message


class MockPackage(object):
    def __init__(self, name, config):
        self.name = name
        self.config = config


class MockPackageSet(object):
    def __init__(self):
        self.installed_packages = None

    def add_packages(self, pkgs):
        if self.installed_packages is None:
            self.installed_packages = []

        for rpm, config in pkgs.items():
            self.installed_packages.append(MockPackage(rpm, config))

    def is_installed(self, pkg):
        for rpm in self.installed_packages:
            if pkg == rpm.name:
                return True
        return False

    def append_content(self, path, content):
        found = False

        for rpm in self.installed_packages:
            if path == rpm.config:
                found = True
        if not found:
            raise IOError('Error during writing to file: {}.'.format(path))


class ExpectedOutput(object):
    def __init__(self):
        self.debugmsg = ''
        self.errmsg = ''

    def create(self, rpms):
        error_list = []

        for pkg, config in rpms.items():
            if pkg in vim_configs.keys():
                self.debugmsg += 'Updating Vim configuration file {}.'.format(vim_configs[pkg])
                if config == '':
                    error_list.append((vim_configs[pkg], 'Error during writing to file: {}.'.format(vim_configs[pkg])))

        if error_list:
            self.errmsg = ('The files below have not been modified '
                           '(error message included):' +
                           ''.join(['\n    - {}: {}'.format(err[0], err[1])
                                   for err in error_list]))


@pytest.mark.parametrize('rpms', packages)
def test_update_vim(rpms):
    logger = MockLogger()
    installed_packages = MockPackageSet()

    installed_packages.add_packages(rpms)

    expected = ExpectedOutput()
    expected.create(rpms)

    update_vim(logger.debug,
               logger.error,
               installed_packages.is_installed,
               installed_packages.append_content)

    assert expected.debugmsg == logger.debugmsg
    assert expected.errmsg == logger.errmsg



from leapp import reporting
from leapp.reporting import create_report


def _report_foreign():
    create_report([
        reporting.Title(
            'device-mapper-multipath now defaults to ignoring foreign devices'
        ),
        reporting.Summary(
            'In RHEL-9, the default value for the "enable_foreign" option has '
            'changed to "NONE". This means that multipath will no longer list '
            'devices that are not managed by device-mapper. In order to retain '
            'the default RHEL-8 behavior of listing foreign multipath devices, '
            '\'enable_foreign ""\' will be added to the defaults section of '
            '"/etc/multipath.conf". If you wish to change to the default '
            'RHEL-9 behavior, please remove this line. This option only '
            'effects the devices that multipath lists. It has no impact on '
            'what devices are managed.'),
        reporting.Severity(reporting.Severity.INFO),
        reporting.Groups([reporting.Groups.SERVICES]),
        reporting.RelatedResource('package', 'device-mapper-multipath')
    ])


def _report_allow_usb():
    create_report([
        reporting.Title(
            'device-mapper-multipath now defaults to ignoring USB devices'
        ),
        reporting.Summary(
            'In RHEL-9, the default multipath configuration has changed to '
            'ignore USB devices. A new config option, "allow_usb_devices" has '
            'been added to control this.  In order to retain the RHEL-8 '
            'behavior of treating USB devices like other block devices. '
            '"allow_usb_devices yes" will be added to the defaults section '
            'of "/etc/multipath.conf". If you wish to change to the default '
            'RHEL-9 behavior, please remove this line.'),
        reporting.Severity(reporting.Severity.INFO),
        reporting.Groups([reporting.Groups.SERVICES]),
        reporting.RelatedResource('package', 'device-mapper-multipath')
    ])


def _create_paths_str(paths):
    if len(paths) < 2:
        return paths[0]
    return '{} and {}'.format(', '.join(paths[0:-1]), paths[-1])


def _report_invalid_regexes(paths):
    paths_str = _create_paths_str(paths)
    create_report([
        reporting.Title(
            'device-mapper-multipath no longer accepts "*" as a valid regular expression'
        ),
        reporting.Summary(
            'Some options in device-mapper-multipath configuration files '
            'have values that are regular expressions. In RHEL-8, if such an '
            'option had a value of "*", multipath would internally convert it '
            'to ".*". In RHEL-9, values of "*" are no longer accepted. '
            'These regular expression values have been found in {}. They '
            'will be converted to ".*"'.format(paths_str)),
        reporting.Severity(reporting.Severity.INFO),
        reporting.Groups([reporting.Groups.SERVICES]),
        reporting.RelatedResource('package', 'device-mapper-multipath')
    ])


def check_configs(facts):
    need_foreign = not any(x for x in facts.configs if x.enable_foreign_exists)
    need_allow_usb = not any(x for x in facts.configs if x.allow_usb_exists)
    invalid_regexes = [x.pathname for x in facts.configs if x.invalid_regexes_exist]

    if need_foreign:
        _report_foreign()
    if need_allow_usb:
        _report_allow_usb()
    if invalid_regexes:
        _report_invalid_regexes(invalid_regexes)



from leapp.actors import Actor
from leapp.libraries.actor import selinuxcontentscanner
from leapp.models import RpmTransactionTasks, SELinuxCustom, SELinuxFacts, SELinuxModules, SELinuxRequestRPMs
from leapp.tags import FactsPhaseTag, IPUWorkflowTag


class SELinuxContentScanner(Actor):
    """
    Scan the system for any SELinux customizations

    Find SELinux policy customizations (custom policy modules and changes
    introduced by semanage) and save them in SELinuxModules and SELinuxCustom
    models. Customizations that are incompatible with SELinux policy on RHEL-8
    are removed.
    """

    name = 'selinuxcontentscanner'
    consumes = (SELinuxFacts,)
    produces = (SELinuxModules, SELinuxCustom, SELinuxRequestRPMs, RpmTransactionTasks)
    tags = (FactsPhaseTag, IPUWorkflowTag)

    def process(self):
        # exit if SELinux is disabled
        for fact in self.consume(SELinuxFacts):
            if not fact.enabled:
                return

        (semodule_list, template_list, rpms_to_install,) = selinuxcontentscanner.get_selinux_modules()

        self.produce(
            SELinuxModules(
                modules=semodule_list,
                templates=template_list
            )
        )
        self.produce(
            RpmTransactionTasks(
                to_install=rpms_to_install
            )
        )
        # this is produced so that we can later verify that the RPMs are present after upgrade
        self.produce(
            SELinuxRequestRPMs(
                to_install=rpms_to_install
            )
        )

        (semanage_valid, semanage_removed,) = selinuxcontentscanner.get_selinux_customizations()

        self.produce(
            SELinuxCustom(
                commands=semanage_valid,
                removed=semanage_removed
            )
        )



import functools

import pytest

from leapp import reporting
from leapp.libraries.actor import nischeck
from leapp.libraries.common.testutils import create_report_mocked, CurrentActorMocked
from leapp.libraries.stdlib import api
from leapp.models import DistributionSignedRPM, NISConfig, RPM

_generate_rpm = functools.partial(RPM,
                                  pgpsig='RSA/SHA256, Mon 01 Jan 1970 00:00:00 AM -03, Key ID 199e2f91fd431d51',
                                  packager='Red Hat, Inc. <http://bugzilla.redhat.com/bugzilla>',
                                  arch='noarch')


@pytest.mark.parametrize('pkgs_installed', [
    ('ypserv',),
    ('ypbind',),
    ('ypserv-not-ypbind',),
    ('ypbind', 'ypserv'),
])
@pytest.mark.parametrize('pkgs_configured', [
    (),
    ('ypbind',),
    ('ypserv',),
    ('ypbind', 'ypserv'),
])
def test_actor_nis(monkeypatch, pkgs_installed, pkgs_configured):
    """
    Parametrized helper function for test_actor_* functions.

    First generate list of RPM models based on set arguments. Then, run
    the actor fed with our RPM list and mocked functions. Finally, assert
    Reports according to set arguments.

    Parameters:
        pkgs_installed  (tuple): installed pkgs
        fill_conf_file  (bool): not default ypbind config file
        fill_ypserv_dir  (bool): not default ypserv dir content
    """

    # Generate few standard packages
    rpms = [_generate_rpm(name='rclone', version='2.3', release='3', epoch='1'),
            _generate_rpm(name='gdb', version='2.0.1', release='1', epoch='2')]

    # Generate packages from 'pkgs_installed'
    for pkg_name in pkgs_installed:
        rpms += [_generate_rpm(name=pkg_name, version='2.0', release='3', epoch='1')]

    # Generate NIS facts
    nis_facts = NISConfig(nis_not_default_conf=pkgs_configured)

    curr_actor_mocked = CurrentActorMocked(msgs=[DistributionSignedRPM(items=rpms), nis_facts])
    monkeypatch.setattr(api, 'current_actor', curr_actor_mocked)
    monkeypatch.setattr(reporting, "create_report", create_report_mocked())

    # Executed actor fed with out fake msgs
    nischeck.report_nis()

    # Iterate through installed packages
    for pkg in pkgs_installed:
        # Check if package is configured
        if pkg in pkgs_configured:
            # Don't waste time checking other conditions
            # if one of pkgs is already found
            assert reporting.create_report.called == 1
            break
    else:
        # Assert for no NIS installed and configured packages
        assert not reporting.create_report.called



from leapp.models import fields, Model
from leapp.topics import SystemInfoTopic


class DaemonList(Model):
    """
    A split up representation of a daemon_list (see host_access(5)). Example value of the
    'value' attribute: ["ALL", "EXCEPT", "in.fingerd"]
    """
    topic = SystemInfoTopic

    value = fields.List(fields.String())


class TcpWrappersFacts(Model):
    """
    A representation of tcp_wrappers configuration. Currently it only contains a list
    of daemon lists that are present in the tcp_wrappers configuration files. From this
    you can extract information on whether there is any configuration that applies to
    a specific daemon (see leapp.libraries.common.tcpwrappersutils.config_applies_to_daemon()).
    """
    topic = SystemInfoTopic

    daemon_lists = fields.List(fields.Model(DaemonList))



from leapp import reporting
from leapp.libraries.stdlib import api


def _normalize_key(key):
    """
    Strip the part of the key before the first dot
    """
    s = key.split(".", 1)
    if len(s) == 2:
        return s[1]
    return key


def _key_equal(pair, key):
    """
    Check the keys are equal in OpenSSL configuration semantics

    The OpenSSL semantics ignores everything before the first dot to allow specifying
    something like following, where the first line would be otherwise normally ignored

        TLS.MaxProtocol = TLSv1.3
        DTLS.MaxProtocol = DTLSv1.2
    """
    if pair.key == key:
        return True
    return _normalize_key(pair.key) == key


def _find_pair(block, key):
    """
    Find key-value pair in the given configuration block

    In the given configuration block (OpenSslConfigBlock) find a key-value with a given key.
    If multiple values match, only the last one is returned.
    """
    res = None
    for pair in block.pairs:
        if _key_equal(pair, key):
            res = pair

    return res


def _openssl_find_block(config, name):
    """
    In the given configuration file (OpenSslConfig) find a block with a given name
    """
    for block in config.blocks:
        if block.name == name:
            return block

    return None


def _openssl_reachable_block(config, start_name, end_name, limit=10):
    """
    Check if the end_name is reachable from the start_name in given config

    This searches path between one block name and some other block name in a chain
    like (the names in square braces are block names, the other keys pointing to another
    block in the chain)
    [default_modules] -> ssl_conf -> [ssl_module] -> system_default -> [crypto_policy]
    using simple recursion. The end block name is not checked for presence so it can
    be just value.
    """
    if limit <= 0:
        api.current_logger().debug("Recursion limit reached while searching for {}."
                                   .format(end_name))
        return False

    if start_name == end_name:
        return True

    start = _openssl_find_block(config, start_name)
    if not start:
        api.current_logger().debug("Starting block {} not found in current configuration"
                                   .format(start_name))
        return False

    for pair in start.pairs:
        if pair.value == end_name:
            return True
        block = _openssl_find_block(config, pair.value)
        if block and _openssl_reachable_block(config, block.name, end_name, limit - 1):
            return True

    return False


def _openssl_reachable_block_root(config, end_name, limit=10):
    """
    Check if the given block is reachable from the root block given in the config.
    """
    return _openssl_reachable_block(config, config.openssl_conf, end_name, limit)


def _openssl_reachable_key(config, key, value=None):
    """
    Check if the key=value pair is reachable from the root block.

    If no value is specified, any value assigned to the given key is matched.
    """
    for block in config.blocks:
        for pair in block.pairs:
            if _key_equal(pair, key) and (value is None or pair.value == value):
                api.current_logger().debug("The key {} found in block {}"
                                           .format(key, block.name))
                if _openssl_reachable_block_root(config, block.name):
                    api.current_logger().debug("The block {} is reachable from the start key {}"
                                               .format(block.name, config.openssl_conf))
                    return True

    api.current_logger().debug("The key {} not found".format(key))
    return False


# pylint: disable=too-many-return-statements -- could not simplify more
def _openssl_reachable_path(config, path, value=None):
    """
    Check if the given path is reachable in OpenSSL configuration

    The path is list where to search for a value. It is a list of keys starting
    with a section of a known name followed by key to search for in that section
    and then another section, which is assigned to that key.

    If no value is specified, any value assigned to the given key is matched.
    """
    path_iterator = iter(path)
    block_name = next(path_iterator)
    if not _openssl_reachable_block_root(config, block_name):
        api.current_logger().debug("The block {} not reachable from the start key {}"
                                   .format(block_name, config.openssl_conf))
        return False

    while block_name and block_name != value:
        # Find in the block by name
        block = _openssl_find_block(config, block_name)
        if not block:
            api.current_logger().debug("Block {} not found in current configuration"
                                       .format(block_name))
            return False

        try:
            key = next(path_iterator)
        except StopIteration:
            api.current_logger().debug("Missing key in the path")
            return False

        # Find the pair in block
        pair = _find_pair(block, key)
        if not pair:
            api.current_logger().debug("Key {} not found in block {}".format(key, block_name))
            return False

        try:
            block_name = next(path_iterator)
        except StopIteration:
            # if there are no other parts in the path and the value is the one we look for, stop
            if pair.value == value or value is None:
                api.current_logger().debug("Found reachable value {}".format(pair.value))
                return True

        # otherwise make sure it is the expected one
        if pair.value != block_name:
            return False

    api.current_logger().debug("Value {} not reachable in current configuration".format(value))
    return False


def _find_duplicates(config, block_name):
    """
    Finds duplicate keys in given block name
    """
    block = _openssl_find_block(config, block_name)
    if not block:
        api.current_logger().debug("Did not find a block {} when searching for duplicates"
                                   .format(block_name))
        return []

    duplicates = []
    # not most effective
    for p in block.pairs:
        for p2 in block.pairs:
            key = _normalize_key(p.key)
            if p != p2 and key == _normalize_key(p2.key) and key not in duplicates:
                duplicates.append(key)

    return duplicates


def _openssl_duplicate_keys_in(config, key):
    """
    Search for duplicate keys in block named "key"

    Used for searching for duplicate extensions. The given key can be defined in different
    blocks and references a block where we search for duplicate keys with the openssl specific
    name handling (ignoring content before the dot).
    """
    duplicates = []

    # first find all the keys in all the blocks with the given name
    for block in config.blocks:
        for pair in block.pairs:
            if _key_equal(pair, key):
                duplicates += _find_duplicates(config, pair.value)

    return duplicates


resources = [
    reporting.RelatedResource('package', 'openssl'),
    reporting.RelatedResource('file', '/etc/pki/tls/openssl.cnf')
]


def check_crypto_policies(config):
    """
    Check the presence of the crypto policies include

    The default openssl.cnf provides include of dynamic crypto policies to keep
    applications using OpenSSL in sync with the rest of the OS. Not having this
    directive in the configuration file is not recommended.
    """
    if not _openssl_reachable_path(config,
                                   path=("default_modules", "ssl_conf", "ssl_module",
                                         "system_default", "crypto_policy", ".include"),
                                   value="/etc/crypto-policies/back-ends/opensslcnf.config"):
        reporting.create_report([
            reporting.Title('The OpenSSL configuration is missing the crypto policies integration'),
            reporting.Summary(
                'The OpenSSL configuration file `/etc/pki/tls/openssl.cnf` does not contain the '
                'directive to include the system-wide crypto policies. This is not recommended '
                'by Red Hat and can lead to decreasing overall system security and inconsistent '
                'behavior between applications. If you need to adjust the crypto policies to your '
                'needs, it is recommended to use custom crypto policies.'
            ),
            reporting.Severity(reporting.Severity.MEDIUM),
            reporting.Groups([
                    reporting.Groups.AUTHENTICATION,
                    reporting.Groups.SECURITY,
                    reporting.Groups.NETWORK,
                    reporting.Groups.SERVICES
            ]),
            reporting.RelatedResource('package', 'crypto-policies'),
        ] + resources)


def check_min_max_protocol(config):
    """
    Check for the MinProtocol and MaxProtocol options

    These options changed their meaning in openssl.cnf so better warn the user
    """
    if _openssl_reachable_key(config, "MinProtocol") or _openssl_reachable_key(config, "MaxProtocol"):
        reporting.create_report([
            reporting.Title('The meaning of MinProtocol and MaxProtocol in openssl.cnf changed'),
            reporting.Summary(
                'The OpenSSL configuration file `/etc/pki/tls/openssl.cnf` contain the '
                'directive MinProtocol or MaxProtocol, which was previously applied to both '
                'TLS and DTLS versions. This is no longer case and if you want to limit both '
                'protocols, you need to add another option, for example for '
                '`MinProtocol TLSv1.2` add the following line `DTLS.MinProtocol = DTLSv1.2`.'
            ),
            reporting.Severity(reporting.Severity.MEDIUM),
            reporting.Groups([
                    reporting.Groups.SECURITY,
                    reporting.Groups.NETWORK,
                    reporting.Groups.SERVICES
            ]),
        ] + resources)


def check_duplicate_extensions(config):
    dup = _openssl_duplicate_keys_in(config, "x509_extensions")
    if dup:
        reporting.create_report([
            reporting.Title('There are duplicate x509 extensions defined in openssl.cnf'),
            reporting.Summary(
                'The OpenSSL configuration file `/etc/pki/tls/openssl.cnf` contains the '
                'following duplicate X509 extensions: {}. With OpenSSL 3.0 only the last defined '
                'will get used. Please, review the configuration file and remove duplicate '
                'extensions to silence this warning.'.format(', '.join(dup))
            ),
            reporting.Severity(reporting.Severity.LOW),
            reporting.Groups([
                    reporting.Groups.SECURITY,
                    reporting.Groups.NETWORK,
                    reporting.Groups.SERVICES
            ]),
        ] + resources)


def check_default_modules(config):
    if config.openssl_conf != "default_modules" or not _openssl_find_block(config, "default_modules"):
        reporting.create_report([
            reporting.Title('Non-standard configuration of openssl.cnf'),
            reporting.Summary(
                'The OpenSSL configuration file `/etc/pki/tls/openssl.cnf` does not contain '
                'expected initialization so it can not be updated to support OpenSSL 3.0 '
                'with the new providers.'
            ),
            reporting.Remediation(
                'The openssl.cnf file needs to contain the following initialization: '
                '`openssl_conf = default_modules` and corresponding `[ default_modules] '
                'block. The `openssl_conf` now contains {} or the `[ default_modules ]` '
                'block is missing. '.format(config.openssl_conf)
            ),
            reporting.Groups([reporting.Groups.INHIBITOR]),
            reporting.Severity(reporting.Severity.HIGH),
            reporting.Groups([
                    reporting.Groups.SECURITY,
                    reporting.Groups.NETWORK,
                    reporting.Groups.SERVICES
            ]),
        ] + resources)
    else:
        reporting.create_report([
            reporting.Title('The OpenSSL configuration will be updated to support OpenSSL 3.0'),
            reporting.Summary(
                'The OpenSSL configuration file `/etc/pki/tls/openssl.cnf` will be updated '
                'to support OpenSSL 3.0 with the new providers. The following changes are '
                'going to be applied:\n'
                ' * Change the initialization:\n'
                '\n'
                ' - openssl_conf = default_modules\n'
                ' + openssl_conf = openssl_init\n'
                '\n'
                ' * Rename the default block and link the providers block:\n'
                '\n'
                ' - [default_modules]\n'
                ' + [openssl_init]\n'
                ' + providers = provider_sect\n'
                '\n'
                ' * Add the providers block:\n'
                '\n'
                ' + [provider_sect]\n'
                ' + default = default_sect\n'
                ' + ##legacy = legacy_sect\n'
                ' + \n'
                ' + [default_sect]\n'
                ' + activate = 1\n'
                ' + \n'
                ' + ##[legacy_sect]\n'
                ' + ##activate = 1\n'
            ),
            reporting.Severity(reporting.Severity.INFO),
            reporting.Groups([
                    reporting.Groups.SECURITY,
                    reporting.Groups.NETWORK,
                    reporting.Groups.SERVICES
            ]),
        ] + resources)




from leapp.actors import Actor
from leapp.libraries.actor import scandasd
from leapp.models import TargetUserSpaceUpgradeTasks, UpgradeInitramfsTasks
from leapp.tags import FactsPhaseTag, IPUWorkflowTag


class ScanDASD(Actor):
    """
    In case of s390x architecture, check whether DASD is used.

    The current check is based just on existence of the /etc/dasd.conf file.
    If it exists, produce UpgradeInitramfsTasks msg to ensure the file
    is available inside the target userspace to be able to generate the
    upgrade init ramdisk correctly.
    """

    name = 'scandasd'
    consumes = ()
    produces = (TargetUserSpaceUpgradeTasks, UpgradeInitramfsTasks)
    tags = (IPUWorkflowTag, FactsPhaseTag)

    def process(self):
        scandasd.process()



import os

from leapp import reporting
from leapp.libraries.common.config import version

COMMON_REPORT_TAGS = [reporting.Groups.SANITY]

related = [reporting.RelatedResource('file', '/etc/os-release')]


def skip_check():
    """ Check if an environment variable was used to skip this actor """
    if os.getenv('LEAPP_DEVEL_SKIP_CHECK_OS_RELEASE'):
        reporting.create_report([
            reporting.Title('Skipped OS release check'),
            reporting.Summary('Source RHEL release check skipped via LEAPP_DEVEL_SKIP_CHECK_OS_RELEASE env var.'),
            reporting.Severity(reporting.Severity.HIGH),
            reporting.Groups(COMMON_REPORT_TAGS)
        ] + related)

        return True
    return False


def check_os_version():
    """ Check the RHEL minor version and inhibit the upgrade if it does not match the supported ones """
    if not version.is_supported_version():
        supported_releases = []
        for rel in version.SUPPORTED_VERSIONS:
            for ver in version.SUPPORTED_VERSIONS[rel]:
                supported_releases.append(rel.upper() + ' ' + ver)
        reporting.create_report([
            reporting.Title(
                'The installed OS version is not supported for the in-place upgrade to the target RHEL version'
            ),
            reporting.Summary(
                'The supported OS releases for the upgrade process:\n'
                ' {}'.format('\n'.join(supported_releases))
            ),
            reporting.Severity(reporting.Severity.HIGH),
            reporting.Groups(COMMON_REPORT_TAGS),
            reporting.Groups([reporting.Groups.INHIBITOR]),
            # we want to set a static Key here because of different Title per path
            reporting.Key('1c7a98849a747ec9890f04bf4321de7280970715')
        ] + related)



from leapp import reporting
from leapp.actors import Actor
from leapp.models import LeftoverPackages, RemovedPackages
from leapp.reporting import create_report, Report
from leapp.tags import IPUWorkflowTag, RPMUpgradePhaseTag


class ReportLeftoverPackages(Actor):
    """
    Collect messages about leftover el7 packages and generate report for users.

    Depending on execution of previous actors, generated report contains information that there are still el7 packages
    present on the system, which makes it unsupported or lists packages that have been removed.
    """

    name = 'report_leftover_packages'
    consumes = (LeftoverPackages, RemovedPackages)
    produces = (Report,)
    tags = (RPMUpgradePhaseTag, IPUWorkflowTag)

    def process(self):
        removed_packages = next(self.consume(RemovedPackages), None)
        leftover_packages = next(self.consume(LeftoverPackages), LeftoverPackages())
        to_remove = ['-'.join([pkg.name, pkg.version, pkg.release]) for pkg in leftover_packages.items]

        if removed_packages:
            title = 'Leftover RHEL 7 packages have been removed'

            if removed_packages.items:
                removed = ['-'.join([pkg.name, pkg.version, pkg.release]) for pkg in removed_packages.items]
                create_report([
                    reporting.Title(title),
                    reporting.Summary('Following packages have been removed:\n{}'.format('\n'.join(removed))),
                    reporting.Severity(reporting.Severity.HIGH),
                    reporting.Groups([reporting.Groups.SANITY]),
                ] + [reporting.RelatedResource('package', pkg.name) for pkg in removed_packages.items])
            else:
                summary = ('Following packages have been removed:\n'
                           '{}\n'
                           'Dependent packages may have been removed as well, please check that you are not missing '
                           'any packages.\n'.format('\n'.join(to_remove)))

                create_report([
                    reporting.Title(title),
                    reporting.Summary(summary),
                    reporting.Severity(reporting.Severity.HIGH),
                    reporting.Groups([reporting.Groups.SANITY]),
                ] + [reporting.RelatedResource('package', pkg.name) for pkg in leftover_packages.items])
            return

        if not leftover_packages.items:
            self.log.info('No leftover packages, skipping...')
            return

        summary = 'Following RHEL 7 packages have not been upgraded:\n{}\n'.format('\n'.join(to_remove))
        summary += 'Please remove these packages to keep your system in supported state.\n'
        create_report([
            reporting.Title('Some RHEL 7 packages have not been upgraded'),
            reporting.Summary(summary),
            reporting.Severity(reporting.Severity.HIGH),
            reporting.Groups([reporting.Groups.SANITY]),
        ] + [reporting.RelatedResource('package', pkg.name) for pkg in leftover_packages.items])



from leapp.actors import Actor
from leapp.libraries.actor.biosdevname import check_biosdevname
from leapp.models import KernelCmdlineArg, PersistentNetNamesFacts
from leapp.tags import FactsPhaseTag, IPUWorkflowTag


class Biosdevname(Actor):
    """
    Enable biosdevname on the target RHEL system if all interfaces on the source RHEL
    system use biosdevname naming scheme and if machine vendor is DELL
    """

    name = 'biosdevname'
    consumes = (PersistentNetNamesFacts,)
    produces = (KernelCmdlineArg,)
    tags = (FactsPhaseTag, IPUWorkflowTag)

    def process(self):
        check_biosdevname()



import pytest

from leapp.libraries.actor import checkremovedenvvars
from leapp.libraries.common.testutils import CurrentActorMocked, produce_mocked
from leapp.libraries.stdlib import api
from leapp.utils.report import is_inhibitor


def test_removed_vars(monkeypatch):
    envars = {'LEAPP_GRUB_DEVICE': '/dev/sda'}
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(envars=envars))
    monkeypatch.setattr(api.current_actor, "produce", produce_mocked())
    checkremovedenvvars.process()
    assert api.current_actor.produce.called == 1
    assert 'LEAPP_GRUB_DEVICE' in api.current_actor.produce.model_instances[0].report['summary']
    assert is_inhibitor(api.current_actor.produce.model_instances[0].report)


def test_no_removed_vars(monkeypatch):
    envars = {'LEAPP_SKIP_RHSM': '1'}
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(envars=envars))
    monkeypatch.setattr(api.current_actor, "produce", produce_mocked())
    checkremovedenvvars.process()
    assert not api.current_actor.produce.called
    assert not api.current_actor.produce.model_instances



import pytest

from leapp import reporting
from leapp.libraries.actor.opensshdeprecateddirectivescheck import inhibit_if_deprecated_directives_used
from leapp.libraries.common.testutils import create_report_mocked, logger_mocked
from leapp.models import OpenSshConfig
from leapp.utils.report import is_inhibitor


def test_inhibit_if_deprecated_directives_used(monkeypatch):
    """Tests whether the upgrade is inhibited when deprecated directives are used in config."""
    created_report = create_report_mocked()
    monkeypatch.setattr(reporting, 'create_report', created_report)

    ssh_config = OpenSshConfig(
        permit_root_login=[],
        deprecated_directives=['ShowPatchLevel']
    )

    inhibit_if_deprecated_directives_used(ssh_config)

    fail_description = 'Report entry was not created when deprecated directive found in the ssh config.'
    assert created_report.called == 1, fail_description

    fail_description = 'Report doesn\'t have information about deprecated directive in the title.'
    assert 'deprecated directive' in created_report.report_fields['title'].lower(), fail_description

    fail_description = 'Report doesn\'t contain the (mocked) deprecated directive present in the config.'
    # The report should have the directive in a preserved form (same as found in configuration)
    assert 'ShowPatchLevel' in created_report.report_fields['summary'], fail_description

    assert created_report.report_fields['severity'] == 'high', 'Report has incorrect severity.'

    fail_description = 'Report should have the inhibition flag set when deprecated directive is present.'
    assert is_inhibitor(created_report.report_fields), fail_description

    assert created_report.report_fields['detail']['remediations'], 'Report should carry some remediation information.'


def test_inhibit_if_deprecated_directives_used_no_deprecated_directives(monkeypatch):
    """Tests whether the upgrade is not inhibited when no deprecated directives are used in config."""
    created_report = create_report_mocked()
    monkeypatch.setattr(reporting, 'create_report', created_report)

    ssh_config = OpenSshConfig(
        permit_root_login=[],
        deprecated_directives=[]
    )

    inhibit_if_deprecated_directives_used(ssh_config)
    assert created_report.called == 0, 'No report should be created if no deprecated directive present in the config.'



from leapp.actors import Actor
from leapp.libraries.actor import removeupgradeartifacts
from leapp.tags import InterimPreparationPhaseTag, IPUWorkflowTag


class RemoveUpgradeArtifacts(Actor):
    """
    Removes artifacts left over by previous leapp runs

    After the upgrade process, there might be some leftover files, which need
    to be cleaned up before running another upgrade.

    Removed artifacts:
    - /root/tmp_leapp_py3/ directory (includes ".leapp_upgrade_failed" flag file)
    """

    name = 'remove_upgrade_artifacts'
    consumes = ()
    produces = ()
    tags = (InterimPreparationPhaseTag, IPUWorkflowTag)

    def process(self):
        removeupgradeartifacts.process()



from leapp.actors import Actor
from leapp.libraries.actor import checkopensslconf
from leapp.models import DistributionSignedRPM, Report, TrackedFilesInfoSource
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class CheckOpenSSLConf(Actor):
    """
    Check whether the openssl configuration and openssl-IBMCA.

    See the report messages for more details. The summary is that since RHEL 8
    it's expected to configure OpenSSL via crypto policies. Also, OpenSSL has
    different versions between major versions of RHEL:
      * RHEL 7: 1.0,
      * RHEL 8: 1.1,
      * RHEL 9: 3.0
    So OpenSSL configuration from older system does not have to be 100%
    compatible with the new system. In some cases, the old configuration could
    make the system inaccessible remotely. So new approach is to ensure the
    upgraded system will use always new default /etc/pki/tls/openssl.cnf
    configuration file (the original one will be backed up if modified by user).

    Similar for OpenSSL-IBMCA, when it's expected to configure it again on
    each newer system.
    """

    name = 'check_openssl_conf'
    consumes = (DistributionSignedRPM, TrackedFilesInfoSource)
    produces = (Report,)
    tags = (IPUWorkflowTag, ChecksPhaseTag)

    def process(self):
        checkopensslconf.process()



import os
import tempfile

import pytest

from leapp.libraries.actor import updateetcsysconfigkernel


# TODO [Artem] could be solved
@pytest.mark.skip(reason='Failing on CI complaining about missing leapp.db fiel')
def test_update_kernel_config(monkeypatch):
    temp = tempfile.NamedTemporaryFile(delete=False)
    with open('tests/files/original') as f:
        data = f.readlines()
        temp.writelines(data)
    temp.close()

    updateetcsysconfigkernel.update_kernel_config(temp.name)

    with open(temp.name) as f:
        result = f.readlines()

    with open('tests/files/expected') as f:
        expected = f.readlines()

    assert result == expected

    os.unlink(temp.name)
    assert not os.path.exists(temp.name)



from leapp.actors import Actor
from leapp.libraries.actor.verifycheckresults import check
from leapp.reporting import Report
from leapp.tags import IPUWorkflowTag, ReportPhaseTag


class VerifyCheckResults(Actor):
    """
    Check all generated results messages and notify user about them.

    A report file containing all messages will be generated, together with log messages displayed
    to the user.
    """

    name = 'verify_check_results'
    consumes = (Report,)
    produces = ()
    tags = (ReportPhaseTag, IPUWorkflowTag)

    def process(self):
        check()



import pytest

from leapp.libraries.actor.cupsscanner import ssl_directive_check

testdata = (
    ('\n', False),
    ('smth\n', False),
    ('#ServerCertificate my.crt\n', False),
    ('#ServerKey my.key\n', False),
    ('ServerCertificate\n', False),
    ('ServerKey\n', False),
    ('ServerKey my.key\n', True),
    ('ServerCertificate my.crt\n', True),
    ('ServerCertificate my.crt\nServerKey my.key\n', True)
)


class MockCupsfiles(object):
    def __init__(self, content):
        self.content = content

    def read(self, path):
        return self.content.splitlines(True)


@pytest.mark.parametrize("content,expected", testdata)
def test_ssl_directive_check(content, expected):
    config = MockCupsfiles(content)

    ret = ssl_directive_check(config.read)

    assert ret == expected



from leapp.exceptions import StopActorExecutionError
from leapp.libraries.common import fetch
from leapp.libraries.common.rpms import get_leapp_packages, LeappComponents
from leapp.libraries.stdlib import api
from leapp.models import DeviceDriverDeprecationData, DeviceDriverDeprecationEntry
from leapp.models.fields import ModelViolationError


def process():
    """
    Loads the device and driver deprecation data and produces a DeviceDriverDeprecationData message with its content.
    It will filter the data on the device_type field, based on the choices set in the StringEnum on the
    DeviceDriverDeprecationEntry model
    """
    # This is how you get the StringEnum choices value, so we can filter based on the model definition
    supported_device_types = set(DeviceDriverDeprecationEntry.device_type.serialize()['choices'])

    data_file_name = 'device_driver_deprecation_data.json'
    # NOTE(pstodulk): load_data_assert raises StopActorExecutionError, see
    # the code for more info. Keeping the handling on the framework in such
    # a case as we have no work to do in such a case here.
    deprecation_data = fetch.load_data_asset(api.current_actor(),
                                             data_file_name,
                                             asset_fulltext_name='Device driver deprecation data',
                                             docs_url='',
                                             docs_title='')

    try:
        api.produce(
            DeviceDriverDeprecationData(
                entries=[
                    DeviceDriverDeprecationEntry(**entry)
                    for entry in deprecation_data['data']
                    if entry.get('device_type') in supported_device_types
                ]
            )
        )
    except (ModelViolationError, ValueError, KeyError, AttributeError, TypeError) as err:
        # For the listed errors, we expect this to happen only when data is malformed
        # or manually updated. Corrupted data in the upstream is discovered
        # prior the merge thanks to testing. So just suggest the restoration
        # of the file.
        msg = 'Invalid device and driver deprecation data: {}'.format(err)
        hint = (
            'This issue is usually caused by manual update of the {lp} file.'
            ' The data inside is either incorrect or old. To restore the original'
            ' {lp} file, remove it and reinstall the following packages: {rpms}'
            .format(
                lp='/etc/leapp/file/device_driver_deprecation_data.json',
                rpms=', '.join(get_leapp_packages(component=LeappComponents.REPOSITORY))
            )
        )
        raise StopActorExecutionError(msg, details={'hint': hint})



from leapp.actors import Actor
from leapp.libraries.stdlib import run
from leapp.models import UdevAdmInfoData
from leapp.tags import FactsPhaseTag, IPUWorkflowTag


class UdevAdmInfo(Actor):
    """
    Produces data exported by the "udevadm info" command.
    """

    name = 'udevadm_info'
    consumes = ()
    produces = (UdevAdmInfoData,)
    tags = (IPUWorkflowTag, FactsPhaseTag,)

    def process(self):
        out = run(['udevadm', 'info', '-e'])['stdout']
        self.produce(UdevAdmInfoData(db=out))



import os

from leapp import reporting
from leapp.libraries.common import rhui
from leapp.libraries.common.config.version import get_source_major_version
from leapp.libraries.common.rpms import has_package
from leapp.libraries.stdlib import api
from leapp.models import FirmwareFacts, HybridImage, InstalledRPM

BIOS_PATH = '/boot/grub2/grubenv'
EFI_PATH = '/boot/efi/EFI/redhat/grubenv'


def is_grubenv_symlink_to_efi():
    """
    Check whether '/boot/grub2/grubenv' is a relative symlink to
    '/boot/efi/EFI/redhat/grubenv'.
    """
    return os.path.islink(BIOS_PATH) and os.path.realpath(BIOS_PATH) == os.path.realpath(EFI_PATH)


def is_azure_agent_installed():
    """Check whether 'WALinuxAgent' package is installed."""
    src_ver_major = get_source_major_version()

    family = rhui.RHUIFamily(rhui.RHUIProvider.AZURE)
    azure_setups = rhui.RHUI_SETUPS.get(family, [])

    agent_pkg = None
    for setup in azure_setups:
        if setup.os_version == src_ver_major:
            agent_pkg = setup.extra_info.get('agent_pkg')
            break

    if not agent_pkg:
        return False

    return has_package(InstalledRPM, agent_pkg)


def is_bios():
    """Check whether system is booted into BIOS"""
    ff = next(api.consume(FirmwareFacts), None)
    return ff and ff.firmware == 'bios'


def check_hybrid_image():
    """Check whether the system is using Azure hybrid image."""
    if all([is_grubenv_symlink_to_efi(), is_azure_agent_installed(), is_bios()]):
        api.produce(HybridImage(detected=True))
        reporting.create_report([
            reporting.Title(
                'Azure hybrid (BIOS/EFI) image detected. "grubenv" symlink will be converted to a regular file'
            ),
            reporting.Summary(
                'Leapp detected the system is running on Azure cloud, booted using BIOS and '
                'the "/boot/grub2/grubenv" file is a symlink to "../efi/EFI/redhat/grubenv". In case of such a '
                'hybrid image scenario GRUB is not able to locate "grubenv" as it is a symlink to different '
                'partition and fails to boot. If the system needs to be run in EFI mode later, please re-create '
                'the relative symlink again.'
            ),
            reporting.Severity(reporting.Severity.HIGH),
            reporting.Groups([reporting.Groups.PUBLIC_CLOUD]),
        ])



from leapp import reporting
from leapp.libraries.stdlib import api
from leapp.models import SystemdServicesTasks

FMT_LIST_SEPARATOR = '\n    - '


def _inhibit_upgrade_with_conflicts(conflicts):
    summary = (
        'The requested states for systemd services on the target system are in conflict.'
        ' The following systemd services were requested to be both enabled and'
        ' disabled on the target system:{}{}'
        .format(FMT_LIST_SEPARATOR, FMT_LIST_SEPARATOR.join(sorted(conflicts)))
    )
    report = [
        reporting.Title('Conflicting requirements of systemd service states'),
        reporting.Summary(summary),
        reporting.Severity(reporting.Severity.HIGH),
        reporting.Groups([reporting.Groups.SANITY]),
        reporting.Groups([reporting.Groups.INHIBITOR]),
    ]
    reporting.create_report(report)


def check_conflicts():
    services_to_enable = set()
    services_to_disable = set()
    for task in api.consume(SystemdServicesTasks):
        services_to_enable.update(task.to_enable)
        services_to_disable.update(task.to_disable)

    conflicts = services_to_enable.intersection(services_to_disable)
    if conflicts:
        _inhibit_upgrade_with_conflicts(conflicts)



from leapp.actors import Actor
from leapp.libraries.actor.sourcebootloaderscanner import scan_source_boot_loader_configuration
from leapp.models import SourceBootLoaderConfiguration
from leapp.tags import FactsPhaseTag, IPUWorkflowTag


class SourceBootLoaderScanner(Actor):
    """
    Scans the boot loader configuration on the source system.
    """

    name = 'source_boot_loader_scanner'
    consumes = ()
    produces = (SourceBootLoaderConfiguration,)
    tags = (FactsPhaseTag, IPUWorkflowTag)

    def process(self):
        scan_source_boot_loader_configuration()



import os
from shutil import copy

from leapp.libraries.stdlib import api
from leapp.models import CupsChangedFeatures

CUPSD_CONF = '/etc/cups/cupsd.conf'
CUPSFILES_CONF = '/etc/cups/cups-files.conf'
"""
CUPS configuration files
"""


class FileOperations(object):
    def readlines(self, path):
        if os.path.exists(path):
            with open(path, 'r') as f:
                return f.readlines()
        else:
            raise IOError('Error when reading file {} - file '
                          'does not exist.'.format(path))

    def write(self, path, mode, content):
        if isinstance(content, list):
            content = ''.join(content)
        with open(path, mode) as f:
            f.write(content)

    def copy_to_ssl(self, oldpath):
        copy(oldpath, '/etc/cups/ssl')


def migrate_digest(op):
    """
    Replaces Digest/BasicDigest for Basic

    :param obj op: file operations object
    """
    try:
        lines = op.readlines(CUPSD_CONF)
    except IOError as error:
        raise IOError(error)

    for line in lines:
        for directive in ['AuthType', 'DefaultAuthType']:
            if line.lstrip().startswith(directive):
                auth_line_value = line.lstrip().lstrip(directive).lstrip()
                for value in ['Digest', 'BasicDigest']:
                    if auth_line_value.startswith(value):
                        lines[lines.index(line)] = '{} Basic\n'.format(directive)

    op.write(CUPSD_CONF, 'w', lines)


def migrate_include(include_files, op):
    """
    Concatenates configuration files and remove lines
    with 'Include' directive

    :param list include_files: list of files which contents will be
    concatenated
    :param obj op: file operations object
    """
    error_list = []
    lines = []
    content = []

    for f in include_files:
        try:
            content = op.readlines(f)
            if f != CUPSD_CONF:
                content = ['\n# added by Leapp\n'] + content
            lines += content
        except IOError as error:
            error_list.append('Include directive: {}'.format(error))

    if error_list:
        return error_list

    for line in lines:
        if line.lstrip().startswith('Include'):
            lines[lines.index(line)] = ''

    op.write(CUPSD_CONF, 'w', lines)

    return None


def move_directives(directives, op):
    """
    Moves the directives from cupsd.conf to cups-files.conf

    :param list directives: list of wanted directives
    :param obj op: file operations object
    """
    try:
        cupsd_lines = op.readlines(CUPSD_CONF)
    except IOError as error:
        raise IOError(error)

    lines_to_move = []
    for line in cupsd_lines:
        for name in directives:
            if line.lstrip().startswith(name):
                lines_to_move.append(line)
                cupsd_lines[cupsd_lines.index(line)] = ''

    op.write(CUPSD_CONF, 'w', cupsd_lines)

    if lines_to_move:
        op.write(CUPSFILES_CONF, 'a',
                 '\n# added by Leapp\n{}'.format(''.join(lines_to_move)))


def migrate_certkey(op):
    """
    Copies the key and the certificate to /etc/cups/ssl if both
    are in different dirs, or sets ServerKeychain value to the dir
    where the key and the certificate are. Removes old directives

    :param list directives: list of wanted directives
    :param obj op: file operations object
    """
    try:
        lines = op.readlines(CUPSFILES_CONF)
    except IOError as error:
        raise IOError(error)

    certkey_values = []

    for line in lines:
        for name in ['ServerKey', 'ServerCertificate']:
            if line.lstrip().startswith(name):
                value = line.split()[1]
                if value.startswith('ssl'):
                    value = os.path.join('/etc/cups', value)
                certkey_values.append(value)
                lines[lines.index(line)] = ''

    op.write(CUPSFILES_CONF, 'w', lines)

    # we need to decide whether we copy the files to /etc/cups/ssl
    # or set ServerKeychain to non-default directory or do nothing
    if all(os.path.dirname(val) == '/etc/cups/ssl' for val in certkey_values):
        return

    # Check that all files are inside the same directory
    if len(set([os.path.dirname(certkey) for certkey in certkey_values])) == 1:
        path = os.path.dirname(certkey_values[0])
        op.write(CUPSFILES_CONF, 'a',
                 '\n# added by Leapp\nServerKeychain {}\n'.format(path))
    else:
        for value in certkey_values:
            if not os.path.dirname(value) == '/etc/cups/ssl':
                op.copy_to_ssl(value)


def _get_facts(model):
    """
    Consumes input data model

    :param class model: name of model which we consume
    """
    return next(api.consume(model), None)


def migrate_configuration(error_log=api.current_logger().error,
                          debug_log=api.current_logger().debug,
                          op=FileOperations(),
                          consume_function=_get_facts):
    """
    Migrate CUPS configuration based on gathered facts

    :param func error_log: sends error messages
    :param func debug_log: sends debug messages
    :param obj op: IO operations
    :param func consume_function: receives data object from a model
    """

    facts = consume_function(CupsChangedFeatures)
    error_list = []

    if not facts:
        return

    if facts.include:
        debug_log('Migrating CUPS configuration - Include directives.')
        include_errors = []

        include_errors = migrate_include(facts.include_files, op)
        if include_errors:
            error_list += include_errors

    if facts.digest:
        debug_log('Migrating CUPS configuration - BasicDigest/Digest directives.')

        try:
            migrate_digest(op)
        except IOError as error:
            error_list.append('Digest/BasicDigest values: {}'.format(error))

    if facts.env:
        debug_log('Migrating CUPS configuration - PassEnv/SetEnv directives.')

        try:
            move_directives(['PassEnv', 'SetEnv'],
                            op)
        except IOError as error:
            error_list.append('PassEnv/SetEnv directives: {}'.format(error))

    if facts.certkey:
        debug_log('Migrating CUPS configuration - '
                  'ServerKey/ServerCertificate directive.')

        try:
            migrate_certkey(op)
        except IOError as error:
            error_list.append('ServerKey/ServerCertificate directives: {}'.format(error))

    if facts.printcap:
        debug_log('Migrating CUPS configuration - PrintcapFormat directive.')

        try:
            move_directives(['PrintcapFormat'],
                            op)
        except IOError as error:
            error_list.append('PrintcapFormat directive: {}'.format(error))

    if error_list:
        error_log('Following errors happened during CUPS migration:'
                  + ''.join(['\n   - {}'.format(err) for err in error_list]))



from leapp.libraries.actor import spamassassinconfigupdate
from leapp.libraries.common.spamassassinutils import SPAMC_CONFIG_FILE, SYSCONFIG_SPAMASSASSIN, SYSCONFIG_VARIABLE
from leapp.models.spamassassinfacts import SpamassassinFacts


class MockBackup(object):
    def __init__(self, to_raise=None):
        self.to_raise = to_raise
        self.called = 0
        self.paths = []

    def __call__(self, path):
        self.called += 1
        self.paths.append(path)
        if self.to_raise:
            raise self.to_raise
        return '/path/to/backup'


class MockFileOperations(object):
    def __init__(self, read_error=None, write_error=None):
        self.files = {}
        self.files_read = {}
        self.files_written = {}
        self.read_called = 0
        self.write_called = 0
        self.read_error = read_error
        self.write_error = write_error

    def _increment_read_counters(self, path):
        self.read_called += 1
        self.files_read.setdefault(path, 0)
        self.files_read[path] += 1

    def read(self, path):
        self._increment_read_counters(path)
        if self.read_error:
            raise self.read_error
        return self.files[path]

    def _increment_write_counters(self, path):
        self.write_called += 1
        self.files_written.setdefault(path, 0)
        self.files_written[path] += 1

    def write(self, path, content):
        self._increment_write_counters(path)
        if self.write_error:
            raise self.write_error
        self.files[path] = content


def test_migrate_configs():
    facts = SpamassassinFacts(spamc_ssl_argument='sslv3',
                              spamd_ssl_version='sslv3',
                              service_overriden=False)
    fileops = MockFileOperations()
    fileops.files[SPAMC_CONFIG_FILE] = '--ssl sslv3\n# foo\n-B\n'
    content = ('# foo\n' +
               SYSCONFIG_VARIABLE + '="-c --ssl-version sslv3 -hdx"\n' +
               '# bar \n')
    fileops.files[SYSCONFIG_SPAMASSASSIN] = content
    backup_func = MockBackup()

    spamassassinconfigupdate.migrate_configs(facts, fileops, backup_func)

    assert backup_func.called == 2
    assert SPAMC_CONFIG_FILE in backup_func.paths
    assert SYSCONFIG_SPAMASSASSIN in backup_func.paths
    assert fileops.files[SPAMC_CONFIG_FILE] == '--ssl\n# foo\n-B\n'
    expected_content = ('# foo\n' +
                        SYSCONFIG_VARIABLE + '="-c --ssl -hx"\n' +
                        '# bar \n')
    assert fileops.files[SYSCONFIG_SPAMASSASSIN] == expected_content



from leapp.exceptions import StopActorExecutionError
from leapp.libraries.common.config import architecture
from leapp.libraries.stdlib import api, CalledProcessError, run
from leapp.models import BootContent, FirmwareFacts


def remove_boot_entry():
    # we need to make sure /boot is mounted before trying to remove the boot entry

    facts_msg = api.consume(FirmwareFacts)
    facts = next(facts_msg, None)
    if not facts:
        raise StopActorExecutionError('Could not identify system firmware',
                                      details={'details': 'Actor did not receive FirmwareFacts message.'})

    mount_points_per_firmware = {
        'bios': ['/boot'],
        'efi': ['/boot', '/boot/efi']
    }

    for mp in mount_points_per_firmware[facts.firmware]:
        try:
            run(['/bin/mount', mp])
        except CalledProcessError:
            # partitions have been most likely already mounted
            pass
    kernel_filepath = get_upgrade_kernel_filepath()
    run([
        '/usr/sbin/grubby',
        '--remove-kernel={0}'.format(kernel_filepath)
    ])
    if architecture.matches_architecture(architecture.ARCH_S390X):
        # on s390x we need to call zipl explicitly because of issue in grubby,
        # otherwise the new boot entry will not be set as default
        # See https://bugzilla.redhat.com/show_bug.cgi?id=1764306
        run(['/usr/sbin/zipl'])

    # TODO: Move calling `mount -a` to a separate actor as it is not really related to removing the upgrade boot entry.
    #       It's worth to call it after removing the boot entry to avoid boot loop in case mounting fails.
    run([
        '/bin/mount', '-a'
    ])


def get_upgrade_kernel_filepath():
    boot_content_msgs = api.consume(BootContent)
    boot_content = next(boot_content_msgs, None)
    if list(boot_content_msgs):
        api.current_logger().warning('Unexpectedly received more than one BootContent message.')
    if not boot_content:
        raise StopActorExecutionError('Could not remove GRUB boot entry for the upgrade initramfs.',
                                      details={'details': 'Did not receive a message about the leapp-provided'
                                                          ' kernel and initramfs.'})
    return boot_content.kernel_path



from leapp import reporting
from leapp.actors import Actor
from leapp.libraries.common.rpms import has_package
from leapp.models import DistributionSignedRPM
from leapp.reporting import create_report, Report
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class CheckAcpid(Actor):
    """
    Check if acpid is installed. If yes, write information about non-compatible changes.
    """

    name = 'checkacpid'
    consumes = (DistributionSignedRPM,)
    produces = (Report,)
    tags = (ChecksPhaseTag, IPUWorkflowTag)

    def process(self):
        if has_package(DistributionSignedRPM, 'acpid'):
            create_report([
                reporting.Title('Acpid incompatible changes in the next major version'),
                reporting.Summary('The option -d (debug) no longer implies -f (foreground).'),
                reporting.Severity(reporting.Severity.LOW),
                reporting.Remediation(
                    hint='You must now use both options (\'-df\') for the same behavior. Please update '
                         'your scripts to be compatible with the changes.'),
                reporting.Groups([reporting.Groups.KERNEL, reporting.Groups.SERVICES]),
                reporting.RelatedResource('package', 'acpid')
            ])



import os

from leapp import reporting
from leapp.libraries.common.config.version import get_source_major_version
from leapp.libraries.common.rhsm import skip_rhsm

# If LEAPP_NO_RHSM is set, subscription-manager and product-id will not be
# considered as required when checking whether the required plugins are enabled.
REQUIRED_YUM_PLUGINS = {'subscription-manager', 'product-id'}
FMT_LIST_SEPARATOR = '\n    - '


def check_required_yum_plugins_enabled(pkg_manager_info):
    """
    Checks whether the yum plugins required by the IPU are enabled.

    If they are not enabled, a report is produced informing the user about it.

    :param pkg_manager_info: PkgManagerInfo
    """

    missing_required_plugins = REQUIRED_YUM_PLUGINS - set(pkg_manager_info.enabled_plugins)

    if skip_rhsm():
        missing_required_plugins -= {'subscription-manager', 'product-id'}

    if missing_required_plugins:
        missing_required_plugins_text = ''
        for missing_plugin in missing_required_plugins:
            missing_required_plugins_text += '{0}{1}'.format(FMT_LIST_SEPARATOR, missing_plugin)

        if get_source_major_version() == '7':
            pkg_manager = 'YUM'
            pkg_manager_config_path = '/etc/yum.conf'
            plugin_configs_dir = '/etc/yum/pluginconf.d'
        else:
            # On RHEL8+ the yum package might not be installed
            pkg_manager = 'DNF'
            pkg_manager_config_path = '/etc/dnf/dnf.conf'
            plugin_configs_dir = '/etc/dnf/plugins'

        # pkg_manager_config_path - enable/disable plugins globally
        # subscription_manager_plugin_conf, product_id_plugin_conf - plugins can be disabled individually
        subscription_manager_plugin_conf = os.path.join(plugin_configs_dir, 'subscription-manager.conf')
        product_id_plugin_conf = os.path.join(plugin_configs_dir, 'product-id.conf')

        remediation_commands = [
            'sed -i \'s/^plugins=0/plugins=1/\' \'{0}\''.format(pkg_manager_config_path),
            'sed -i \'s/^enabled=0/enabled=1/\' \'{0}\''.format(subscription_manager_plugin_conf),
            'sed -i \'s/^enabled=0/enabled=1/\' \'{0}\''.format(product_id_plugin_conf)
        ]

        reporting.create_report([
            reporting.Title('Required {0} plugins are not being loaded.'.format(pkg_manager)),
            reporting.Summary(
                'The following {0} plugins are not being loaded: {1}'.format(pkg_manager,
                                                                             missing_required_plugins_text)
            ),
            reporting.Remediation(
                hint='If you have yum plugins globally disabled, please enable them by editing the {0}. '
                     'Individually, the {1} plugins can be enabled in their corresponding configurations found at: {2}'
                     .format(pkg_manager_config_path, pkg_manager, plugin_configs_dir),
                # Provide all commands as one due to problems with satellites
                commands=[['bash', '-c', '"{0}"'.format('; '.join(remediation_commands))]]
            ),
            reporting.RelatedResource('file', pkg_manager_config_path),
            reporting.RelatedResource('file', subscription_manager_plugin_conf),
            reporting.RelatedResource('file', product_id_plugin_conf),
            reporting.Severity(reporting.Severity.HIGH),
            reporting.Groups([reporting.Groups.INHIBITOR]),
            reporting.Groups([reporting.Groups.REPOSITORY]),
        ])



from leapp.actors import Actor
from leapp.libraries.actor.checkbootavailspace import (
    check_avail_space_on_boot,
    get_avail_bytes_on_boot,
)
from leapp.reporting import Report
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class CheckBootAvailSpace(Actor):
    """
    Check if at least 100Mib of available space on /boot. If not, inhibit the upgrade process.

    Rationale for the requirement of 100MiB:
    - Before reboot into initramfs, the CopyInitramfsToBoot actor copies kernel and initramfs to
      /boot, together worth of 66MiB.
    - After booting into initramfs, the RemoveBootFiles actor removes the copied kernel and
      initramfs from /boot.
    - The DnfShellRpmUpgrade installs a new kernel-core package which puts additional 54MiB of data
      to /boot.
    - Even though the available space needed at the time of writing this actor is 66MiB, the
      additional 100-66=34MiB is a leeway for potential growth of the kernel or initramfs in size.
    """

    name = 'check_boot_avail_space'
    consumes = ()
    produces = (Report,)
    tags = (IPUWorkflowTag, ChecksPhaseTag)

    def process(self):
        check_avail_space_on_boot(get_avail_bytes_on_boot)



import pytest

from leapp.libraries.common.config import version
from leapp.models import IpaInfo
from leapp.reporting import Report
from leapp.snactor.fixture import current_actor_context
from leapp.utils.report import is_inhibitor


def mock_ipa_info(client, server_pkg, server_configured):
    return IpaInfo(
        has_client_package=client,
        is_client_configured=client,
        has_server_package=server_pkg,
        is_server_configured=server_configured,
    )


@pytest.mark.parametrize('src_v', ['7', '8'])
def test_inhibit_ipa_configured(monkeypatch, current_actor_context, src_v):
    monkeypatch.setattr(version, "get_source_major_version", lambda: src_v)
    current_actor_context.feed(mock_ipa_info(True, True, True))
    current_actor_context.run()
    reports = current_actor_context.consume(Report)

    assert len(reports) == 1
    fields = reports[0].report
    assert is_inhibitor(fields)
    assert "ipa-server" in fields["title"]


@pytest.mark.parametrize('src_v', ['7', '8'])
def test_warn_server_pkg(monkeypatch, current_actor_context, src_v):
    monkeypatch.setattr(version, "get_source_major_version", lambda: src_v)
    current_actor_context.feed(mock_ipa_info(True, True, False))
    current_actor_context.run()
    reports = current_actor_context.consume(Report)

    assert len(reports) == 1
    fields = reports[0].report
    assert not is_inhibitor(fields)
    assert "ipa-server" in fields["title"]


def test_client_only(current_actor_context):
    current_actor_context.feed(mock_ipa_info(True, False, False))
    current_actor_context.run()
    reports = current_actor_context.consume(Report)

    assert not reports


def test_no_ipa(current_actor_context):
    current_actor_context.feed(mock_ipa_info(False, False, False))
    current_actor_context.run()
    reports = current_actor_context.consume(Report)

    assert not reports



from leapp import reporting
from leapp.actors import Actor
from leapp.dialogs import Dialog
from leapp.dialogs.components import BooleanComponent
from leapp.models import RemovedPAMModules
from leapp.reporting import create_report, Report
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class RemoveOldPAMModulesCheck(Actor):
    """
    Check if it is all right to disable PAM modules that are not in RHEL-8.

    If admin will refuse to disable these modules (pam_pkcs11 and pam_krb5),
    upgrade will be stopped. Otherwise we would risk locking out the system
    once these modules are removed.
    """
    name = 'removed_pam_modules_check'
    consumes = (RemovedPAMModules,)
    produces = (Report,)
    tags = (IPUWorkflowTag, ChecksPhaseTag)
    dialogs = (
        Dialog(
            scope='remove_pam_pkcs11_module_check',
            reason='Confirmation',
            components=(
                BooleanComponent(
                    key='confirm',
                    label='Disable pam_pkcs11 module in PAM configuration? '
                          'If no, the upgrade process will be interrupted.',
                    description='PAM module pam_pkcs11 is no longer available '
                                'in RHEL-8 since it was replaced by SSSD.',
                    reason='Leaving this module in PAM configuration may '
                           'lock out the system.'
                ),
            )
        ),
        Dialog(
            scope='remove_pam_krb5_module_check',
            reason='Confirmation',
            components=(
                BooleanComponent(
                    key='confirm',
                    label='Disable pam_krb5 module in PAM configuration? '
                          'If no, the upgrade process will be interrupted.',
                    description='PAM module pam_krb5 is no longer available '
                                'in RHEL-8 since it was replaced by SSSD.',
                    reason='Leaving this module in PAM configuration may '
                           'lock out the system.'
                ),
            )
        ),
    )

    modules = []

    def process(self):
        model = next(self.consume(RemovedPAMModules))

        for module in model.modules:
            result = self.confirm(module)
            if result:
                self.produce_report(module)
            elif result is False:
                # user specifically chose to disagree with auto disablement
                self.produce_inhibitor(module)

    def confirm(self, module):
        questions = {
            'pam_pkcs11': self.dialogs[0],
            'pam_krb5': self.dialogs[1]
        }

        return self.get_answers(questions[module]).get('confirm')

    def produce_report(self, module):
        create_report([
            reporting.Title('Module {0} will be removed from PAM configuration'.format(module)),
            reporting.Summary(
                'Module {0} was surpassed by SSSD and therefore it was '
                'removed from RHEL-8. Keeping it in PAM configuration may '
                'lock out the system thus it will be automatically removed '
                'from PAM configuration before upgrading to RHEL-8. '
                'Please switch to SSSD to recover the functionality '
                'of {0}.'.format(module)
            ),
            reporting.Severity(reporting.Severity.MEDIUM),
            reporting.Groups([
                    reporting.Groups.AUTHENTICATION,
                    reporting.Groups.SECURITY,
                    reporting.Groups.TOOLS
            ]),
            reporting.Remediation(hint='Configure SSSD to replace {0}'.format(module)),
            reporting.RelatedResource('package', 'sssd')
        ])

    def produce_inhibitor(self, module):
        create_report([
            reporting.Title(
                'Upgrade process was interrupted because {0} is enabled in '
                'PAM configuration and SA user refused to disable it '
                'automatically.'.format(module)),
            reporting.Summary(
                'Module {0} was surpassed by SSSD and therefore it was '
                'removed from RHEL-8. Keeping it in PAM configuration may '
                'lock out the system thus it is necessary to disable it '
                'before the upgrade process can continue.'.format(module)
            ),
            reporting.Severity(reporting.Severity.HIGH),
            reporting.Groups([
                    reporting.Groups.AUTHENTICATION,
                    reporting.Groups.SECURITY,
                    reporting.Groups.TOOLS
            ]),
            reporting.Groups([reporting.Groups.INHIBITOR]),
            reporting.Remediation(
                hint='Disable {0} module and switch to SSSD to recover its functionality.'.format(module)),
            reporting.RelatedResource('package', 'sssd')
        ])



from leapp.actors import Actor
from leapp.libraries.actor import migrateopensslconf
from leapp.tags import ApplicationsPhaseTag, IPUWorkflowTag


class MigrateOpenSslConf(Actor):
    """
    Enforce the target default configuration file to be used.

    If the /etc/pki/tls/openssl.cnf has been modified and openssl.cnf.rpmnew
    file is created, backup the original one and replace it by the new default.

    tl;dr: (simplified)
    if the file is modified; then
      mv /etc/pki/tls/openssl.cnf{,.leappsave}
      mv /etc/pki/tls/openssl.cnf{.rpmnew,}
    fi
    """

    name = 'migrate_openssl_conf'
    consumes = ()
    produces = ()
    tags = (IPUWorkflowTag, ApplicationsPhaseTag)

    def process(self):
        migrateopensslconf.process()



from leapp.libraries.common.config import get_target_product_channel
from leapp.libraries.common.config.version import get_source_major_version, get_target_major_version
from leapp.libraries.stdlib import api

DEFAULT_PESID = {
    '7': 'rhel7-base',
    '8': 'rhel8-BaseOS',
    '9': 'rhel9-BaseOS'
}


def _get_channel_prio(pesid_repo):
    priorities = {
        'beta': 0,
        'ga': 1,
    }
    return priorities.get(pesid_repo.channel, 10)


class RepoMapDataHandler(object):
    """
    Provide the basic functionality to work with the repository data easily.
    """

    def __init__(self, repo_map, cloud_provider='', default_channels=None):
        """
        Initialize the object based on the given RepositoriesMapping msg.

        Expects that msg contains just stuff related for the current IPU
        (at least mapping and repos for the used upgrade path and architecture).

        :param repo_map: A valid RepositoryMapping message.
        :type repo_map: RepositoryMapping
        :param default_channels: A list of default channels to use when a target repository
                                 equivalent exactly matching a source repository was not found.
        :type default_channels: List[str]
        :param prio_channel: Prefer repositories with this channel when looking for target equivalents.
        :type prio_channel: str
        """
        # NOTE: currently I am keeping this default data structure that is not
        # ideal for work, but there is not any significant impact..
        self.repositories = repo_map.repositories
        self.mapping = repo_map.mapping
        # FIXME(pstodulk): what about default_channel -> fallback_channel
        # hardcoded always as ga? instead of list of channels..
        # it'd be possibly confusing naming now...
        self.default_channels = default_channels or ['ga']

        # Make self.prio_channel None if the user did not specify any target channels, so that self.default_channels
        # will be used instead
        self.prio_channel = get_target_product_channel(default=None)

        self.cloud_provider = cloud_provider

        # Cloud provider might have multiple variants, e.g, aws: (aws, aws-sap-es4) - normalize it
        cloud_providers = ('aws', 'azure', 'google', 'alibaba')
        for provider in cloud_providers:
            if cloud_provider.startswith(provider):
                self.cloud_provider = provider
                break

    def set_default_channels(self, default_channels):
        """
        Set the default channels that are used as a fallback when searching
        for the right target repository.

        Usually it's not problem to find a target repository that matches
        the source repository, however in some cases the target repository
        doesn't have to be available in the required (premium) channel but
        could be present in the standard one.

        This is used usually for fallbacks and it's prerequisite for the time
        the required channel will not be present for the particular repository.
        E.g. can happen for layered products which has different lifecycles
        and doesn't have to provide a special premium channels at all. E.g.
        the Extras repository has GA and Beta channels, but no EUS. In case the
        EUS is required, returns the GA one instead when this is present.

        It's recommended to make the GA ('ga') always present in the default
        list.

        :param default_channels: Default channels to use when a target equivalent to a source repository that
                                 matches its target channel properties exactly could not be found.
        :type default_channels: List[str]
        """
        self.default_channels = default_channels

    def get_pesid_repo_entry(self, repoid, major_version):
        """
        Retrieve the PESIDRepositoryEntry that matches the given repoid and OS major version.

        If multiple pesid repo entries with the same repoid were found, the entry with rhui matching the source
        system's rhui info will be returned. If no entry with matching rhui exists, the CDN one is returned if any.

        :param repoid: RepoID that should the PESIDRepositoryEntry match.
        :type repoid: str
        :param major_version: RepoID that should the PESIDRepositoryEntry match.
        :type major_version: str
        :return: The PESIDRepositoryEntry matching the given repoid and major_version or None if no such
                 entry could be found.
        :rtype: Optional[PESIDRepositoryEntry]
        """
        matching_pesid_repos = []
        for pesid_repo in self.repositories:
            if pesid_repo.repoid == repoid and pesid_repo.major_version == major_version:
                matching_pesid_repos.append(pesid_repo)

        if len(matching_pesid_repos) == 1:
            # Perform no heuristics if only a single pesid repository with matching repoid found
            return matching_pesid_repos[0]

        # Multiple (different) repositories with the same repoid found (can happen in clouds) - prefer the cloud one
        cdn_pesid_repo = None
        for pesid_repo in matching_pesid_repos:
            if pesid_repo.rhui == self.cloud_provider:
                return pesid_repo
            if not pesid_repo.rhui:
                cdn_pesid_repo = pesid_repo

        # If we did not find a repoid for the current cloud provider, return the CDN repository
        return cdn_pesid_repo  # might be None e.g. if we are on Azure with an AWS repository enabled (unlikely)

    def get_target_pesids(self, source_pesid):
        """
        Return sorted list of target PES IDs for the given source PES ID.

        :param source_pesid: Source PES ID to find equivalents for.
        :type source_pesid: src
        :return: The list of target PES IDs the provided source_pesid is mapped to.
        :rtype: List[PESIDRepositoryEntry]
        """
        pesids = set()
        for repomap in self.mapping:
            if repomap.source == source_pesid:
                pesids.update(repomap.target)
        return sorted(pesids)

    def get_pesid_repos(self, pesid, major_version):
        """
        Get the list of PESIDRepositoryEntry with the specified PES ID and OS major version.

        :param pesid: PES ID of the repositories to be retrieved.
        :type pesid: str
        :param major_version: OS major version of the repositories to be retrieved.
        :type major_version: str
        :return: A list of PESIDRepositoryEntries that match the provided PES ID and OS major version.
        :rtype: List[PESIDRepositoryEntry]
        """
        pesid_repos = []
        for pesid_repo in self.repositories:
            if pesid_repo.pesid == pesid and pesid_repo.major_version == major_version:
                pesid_repos.append(pesid_repo)
        return pesid_repos

    def get_source_pesid_repos(self, pesid):
        """
        Return the list of PESIDRepositoryEntry objects for a specified PES ID
        matching the source OS major version.

        :param pesid: The PES ID for which to retrieve PESIDRepositoryEntries.
        :type pesid: str
        :return: A list of PESIDRepositoryEntries that match the provided PES ID and have
                 the OS Major version same as the source OS.
        :rtype: List[PESIDRepositoryEntry]
        """
        return self.get_pesid_repos(pesid, get_source_major_version())

    def get_target_pesid_repos(self, pesid):
        """
        Return the list of PESIDRepositoryEntry objects for a specified PES ID
        matching the target OS major version.

        :param pesid: The PES ID for which to retrieve PESIDRepositoryEntries.
        :type pesid: str
        :return: A list of PESIDRepositoryEntries that match the provided PES ID and have
                 the OS Major version same as the target OS.
        :rtype: List[PESIDRepositoryEntry]
        """
        return self.get_pesid_repos(pesid, get_target_major_version())

    def _find_repository_target_equivalent(self, src_pesidrepo, target_pesid):
        """
        Find the target repository that is the best-match to the source one with the given
        target PES ID.

        :param src_pesidrepo: The source repository to find equivalent to.
        :type src_pesidrepo: PESIDRepositoryEntry
        :param target_pesid: The target PES ID which the target repository must contain.
        :type target_pesid: str
        :return: A target equivalent of given repository.
        :rtype: Optional[PESIDRepositoryEntry]
        """

        candidates = []
        for candidate in self.get_target_pesid_repos(target_pesid):
            matches_rhui = candidate.rhui == src_pesidrepo.rhui
            matches_repo_type = candidate.repo_type == 'rpm'
            matches_arch = candidate.arch == api.current_actor().configuration.architecture

            if matches_rhui and matches_arch and matches_repo_type:
                # user can specify in future the specific channel should be
                # prioritized always (e.g. want to go to EUS...).
                channel = self.prio_channel or src_pesidrepo.channel
                if candidate.channel == channel:
                    return candidate
                candidates.append(candidate)

        # Fallback...
        # Could not find exact-match, so go through candidates if we find an
        # alternative in one of default channels (usually just 'ga')
        for channel in self.default_channels:
            for candidate in candidates:
                if channel == candidate.channel:
                    return candidate

        # This is a case, that must be handled by the caller
        return None

    def get_mapped_target_pesid_repos(self, src_pesidrepo):
        """
        Returns the dict of form {target_pesid: target PESIDRepositoryEntry} containing
        pesids of repositories that is the source pesidrepo mapped to as keys and with
        the best fitting repository for each of the target pesids as values.

        The function always returns the best candidate for every target pesid
        that fits to the given source repository. In case no repository is find
        for a pesid, the value in dict is None: {pesid: None}

        :param src_pesidrepo: The PESIDRepositoryEntry to find the target pesids and the corresponding
                              best-fitting repositories to.
        :type src_pesidrepo: PESIDRepositoryEntry
        :return: A dictionary of the form described above.
        :rtype: Dict[str, PESIDRepositoryEntry]
        """
        result = {}
        for target_pesid in self.get_target_pesids(src_pesidrepo.pesid):
            result[target_pesid] = self._find_repository_target_equivalent(src_pesidrepo, target_pesid)
        return result

    def get_mapped_target_repoids(self, src_pesidrepo):
        """
        Return the list of target repoids for the given src_pesidrepo.

        Some actors do not has to check whether a target repository has been
        found for each target PES ID and details about the target repositories.
        For such actors it's ok to see just list of repoids they should be
        interested about and keep the proper checks for the right actor.

        :param src_pesidrepo: The PESIDRepositoryEntry to find the corresponding best-fitting repositories to.
        :type src_pesidrepo: PESIDRepositoryEntry
        :return: A dictionary of the form described above.
        :rtype: Dict[str, PESIDRepositoryEntry]
        """
        return [repo.repoid for repo in self.get_mapped_target_pesid_repos(src_pesidrepo).values() if repo]

    def get_expected_target_pesid_repos(self, src_repoids):
        """
        Return {target_pesid: PESIDRepositoryEntry} with expected target repositories.

        If some repositories are mapped to a target pesid for which no equivalent
        repository is discovered, such a key contains just None value.

        :param src_repoids: list of present source repoids that should be mapped to target repositories
        :type src_repoids: List[str]
        :rtype: {str: PESIDRepositoryEntry}
        """
        # {pesid: target_repo}
        target_repos_best_candidates = {}
        for src_repoid in src_repoids:
            src_pesidrepo = self.get_pesid_repo_entry(src_repoid, get_source_major_version())
            if not src_pesidrepo:
                # unmapped or custom repo -> skip this one
                continue

            for target_pesid, target_candidate in self.get_mapped_target_pesid_repos(src_pesidrepo).items():
                best_candidate = target_repos_best_candidates.get(target_pesid, None)
                if not best_candidate:
                    # we need to initialize the pesid even when the target_candidate is empty
                    # to know we possibly miss something
                    target_repos_best_candidates[target_pesid] = target_candidate
                if not target_candidate:
                    # It's not crucial in this moment - the pesid can be still filled
                    # by other maps. However log the warning as it is still unexpected
                    # with the valid repomap data
                    api.current_logger().warning(
                        'Cannot find any mapped target repository from the'
                        ' {pesid} family for the {repoid} repository.'
                        .format(repoid=src_repoid, pesid=target_pesid)
                    )
                    continue

                if best_candidate and _get_channel_prio(target_candidate) > _get_channel_prio(best_candidate):
                    # NOTE(pstodulk): we want just one target repository from the "PESID family"
                    # priority: beta < ga < all_else
                    # Why all_else?
                    # -> we do not expect multiple different premium channels present on the system
                    target_repos_best_candidates[target_pesid] = target_candidate
        return target_repos_best_candidates


def get_default_repository_channels(repomap, src_repoids):
    """
    Returns the default repository channels. The 'ga' channel is always included and is the last
    one in the list, so it is the lowest priority when checking for channels.

    :param repomap: A RepoMapDataHandler instance containing the RepositoriesMapping data.
    :type repomap: RepoMapDataHandler
    :param src_repoids: Repositories present on the source system.
    :type src_repoids: List[str]
    :rtype: List[str]
    """
    default_pesid = DEFAULT_PESID[get_source_major_version()]
    top_prio_pesid_repo = None
    for repoid in src_repoids:
        pesid_repo = repomap.get_pesid_repo_entry(repoid, get_source_major_version())
        if not pesid_repo or pesid_repo.pesid != default_pesid:
            continue
        if not top_prio_pesid_repo or _get_channel_prio(pesid_repo) > _get_channel_prio(top_prio_pesid_repo):
            top_prio_pesid_repo = pesid_repo

    # always return at least 'ga'
    if not top_prio_pesid_repo or top_prio_pesid_repo.channel == 'ga':
        return ['ga']

    # keep this order to prefer higher prio to check first
    return [top_prio_pesid_repo.channel, 'ga']



import os.path

from leapp import reporting
from leapp.reporting import create_report

COROSYNC_CONF_LOCATION = "/etc/corosync/corosync.conf"
CIB_LOCATION = "/var/lib/pacemaker/cib/cib.xml"


def inhibit(node_type):
    create_report([
        reporting.Title("Use of HA cluster detected. Upgrade can't proceed."),
        reporting.Summary(
            "HA cluster is not supported by the inplace upgrade.\n"
            "HA cluster configuration file(s) found."
            " It seems to be a cluster {0}.".format(node_type)
        ),
        reporting.Severity(reporting.Severity.HIGH),
        reporting.Groups([reporting.Groups.HIGH_AVAILABILITY]),
        reporting.Groups([reporting.Groups.INHIBITOR]),
        reporting.ExternalLink(
            url="https://access.redhat.com/articles/2059253",
            title=(
                "Recommended Practices for Applying Software Updates"
                " to a RHEL High Availability or Resilient Storage Cluster"
            ),
        ),
        reporting.Remediation(
            hint=(
                "Destroy the existing HA cluster"
                " or (if you have already removed HA cluster packages) remove"
                " configuration files {0} and {1}".format(
                    CIB_LOCATION,
                    COROSYNC_CONF_LOCATION,
                )
            ),
            commands=[["sh", "-c", "pcs cluster stop --all --wait && pcs cluster destroy --all"]]
        ),
        reporting.RelatedResource('file', COROSYNC_CONF_LOCATION),
        reporting.RelatedResource('file', CIB_LOCATION)
    ])


def check_ha_cluster():
    if os.path.isfile(COROSYNC_CONF_LOCATION):
        inhibit(node_type="node")
    elif os.path.isfile(CIB_LOCATION):
        inhibit(node_type="remote node")



import os

from leapp.libraries.actor import networkmanagerreadconfig

CUR_DIR = os.path.dirname(os.path.abspath(__file__))


def test_nm_with_dhcp():
    config = networkmanagerreadconfig.read_nm_config(file_path=os.path.join(CUR_DIR, 'files/nm_cfg_with_dhcp'))
    parser = networkmanagerreadconfig.parse_nm_config(config)

    assert config
    assert parser
    assert parser.has_option('main', 'dhcp')


def test_nm_without_dhcp():
    config = networkmanagerreadconfig.read_nm_config(file_path=os.path.join(CUR_DIR, 'files/nm_cfg_without_dhcp'))
    parser = networkmanagerreadconfig.parse_nm_config(config)

    assert config
    assert parser
    assert not parser.has_option('main', 'dhcp')


def test_nm_with_error():
    config = networkmanagerreadconfig.read_nm_config(file_path=os.path.join(CUR_DIR, 'files/nm_cfg_file_error'))
    parser = networkmanagerreadconfig.parse_nm_config(config)

    assert config
    assert parser
    assert not parser.has_section('main')



import os

from leapp.libraries.stdlib import api, CalledProcessError, run

UPGRADE_ARTIFACTS_DIR = '/root/tmp_leapp_py3/'


def process():
    if os.path.exists(UPGRADE_ARTIFACTS_DIR):
        api.current_logger().debug(
                "Removing leftover upgrade artifacts dir: {} ".format(UPGRADE_ARTIFACTS_DIR))

        try:
            run(['rm', '-rf', UPGRADE_ARTIFACTS_DIR])
        except (CalledProcessError, OSError) as e:
            api.current_logger().debug(
                    'Failed to remove leftover upgrade artifacts dir: {}'.format(e))



from leapp.models import fields, Model
from leapp.topics import SystemInfoTopic


class QuaggaToFrrFacts(Model):
    """
    Model for quagga to frr actors.

    A list of configuration files used by quagga. This list is used to add yes/no to
    /etc/frr/daemons file. It indicates which daemons from frr should be run.
    """
    topic = SystemInfoTopic

    active_daemons = fields.List(fields.String())
    enabled_daemons = fields.List(fields.String())



import os

from leapp.actors import Actor
from leapp.libraries.common.pam import PAM
from leapp.libraries.stdlib import api
from leapp.models import PamConfiguration, PamService
from leapp.tags import FactsPhaseTag, IPUWorkflowTag


class PamModulesScanner(Actor):
    """
    Scan the pam directory for services and modules used in them

    This produces a PAMConfiguration message containing the whole
    list of configured PAM services and what modules they contain.
    """

    name = 'pam_modules_scanner'
    consumes = ()
    produces = (PamConfiguration, )
    tags = (FactsPhaseTag, IPUWorkflowTag, )

    def process(self):
        conf = []
        path = "/etc/pam.d/"
        for f in os.listdir(path):
            pam_file = os.path.join(path, f)
            # Ignore symlinks (usually handled by authconfig)
            if not os.path.isfile(pam_file) or os.path.islink(pam_file):
                continue

            # Use the existing PAM library to parse the files, but unpack it to our model
            try:
                content = PAM.read_file(pam_file)
                modules = PAM(content)
                service = PamService(service=f, modules=modules.modules)
                conf.append(service)
            except OSError as err:
                # if leapp can not read that file it will probably not be important
                api.current_logger().warning('Failed to read file {}: {}'.format(pam_file, err.strerror))

        pam = PamConfiguration(services=conf)
        self.produce(pam)



from leapp.actors import Actor
from leapp.libraries.actor.networkmanagerreadconfig import check_nm_dhcp, parse_nm_config, read_nm_config
from leapp.models import NetworkManagerConfig
from leapp.tags import FactsPhaseTag, IPUWorkflowTag


class NetworkManagerReadConfig(Actor):
    """
    Provides data about NetworkManager configuration.

    After collecting data from NetworkManager tool, a message with relevant data will be produced.
    """

    name = 'network_manager_read_config'
    consumes = ()
    produces = (NetworkManagerConfig,)
    tags = (IPUWorkflowTag, FactsPhaseTag,)

    def process(self):
        nm_config = NetworkManagerConfig()

        cfg = read_nm_config()
        parser = parse_nm_config(cfg)

        if parser:
            check_nm_dhcp(nm_cfg=nm_config, parser=parser)

            self.produce(nm_config)



import pytest

from leapp.exceptions import StopActorExecutionError
from leapp.libraries.actor import opensshalgorithmscheck
from leapp.models import OpenSshConfig, OpenSshPermitRootLogin, Report
from leapp.snactor.fixture import current_actor_context


def test_no_config(current_actor_context):
    with pytest.raises(StopActorExecutionError):
        opensshalgorithmscheck.process(iter([]))


osprl = OpenSshPermitRootLogin(value='no')


@pytest.mark.parametrize('ciphers,expected_report', [
    (None, False),
    ('aes128-ctr', False),
    ('aes128-ctr,aes192-ctr,aes256-ctr', False),
    ('arcfour', True),
    ('arcfour,arcfour128,arcfour256', True),
    ('arcfour,aes128-ctr', True),
    ('aes128-ctr,arcfour', True)])
def test_ciphers(current_actor_context, ciphers, expected_report):
    current_actor_context.feed(OpenSshConfig(
        permit_root_login=[osprl],
        deprecated_directives=[],
        ciphers=ciphers
    ))
    current_actor_context.run()
    if expected_report:
        assert current_actor_context.consume(Report)
    else:
        assert not current_actor_context.consume(Report)



from leapp.actors import Actor
from leapp.libraries.actor import updategrubcore
from leapp.models import FirmwareFacts, TransactionCompleted
from leapp.reporting import Report
from leapp.tags import IPUWorkflowTag, RPMUpgradePhaseTag


class UpdateGrubCore(Actor):
    """
    Update GRUB2 core on legacy BIOS systems.

    On legacy (BIOS) systems, GRUB core (located in the gap between the MBR and the
    first partition), does not get automatically updated when GRUB is upgraded.
    """

    name = 'update_grub_core'
    consumes = (TransactionCompleted, FirmwareFacts)
    produces = (Report,)
    tags = (RPMUpgradePhaseTag, IPUWorkflowTag)

    def process(self):
        updategrubcore.process()



from leapp.models import fields, Model, Module
from leapp.topics import TransactionTopic


class RpmTransactionTasks(Model):
    topic = TransactionTopic

    local_rpms = fields.List(fields.String(), default=[])
    to_install = fields.List(fields.String(), default=[])
    to_keep = fields.List(fields.String(), default=[])
    to_remove = fields.List(fields.String(), default=[])
    to_upgrade = fields.List(fields.String(), default=[])
    modules_to_enable = fields.List(fields.Model(Module), default=[])
    modules_to_reset = fields.List(fields.Model(Module), default=[])


class FilteredRpmTransactionTasks(RpmTransactionTasks):
    pass


class PESRpmTransactionTasks(RpmTransactionTasks):
    # Introduced because the framework struggles with solving dependency order of actors:
    # https://github.com/oamg/leapp/issues/491
    pass



from leapp import tags
from leapp.models import IPUConfig
from leapp.workflows import Workflow
from leapp.workflows.flags import Flags
from leapp.workflows.phases import Phase
from leapp.workflows.policies import Policies
from leapp.workflows.tagfilters import TagFilter


class IPUWorkflow(Workflow):
    """In-Place Upgrade workflow used by the leapp utility to process the in-place upgrade."""

    name = 'InplaceUpgrade'
    tag = tags.IPUWorkflowTag
    short_name = 'ipu'
    configuration = IPUConfig
    description = """The IPU workflow takes care of an in-place upgrade (IPU) of RHEL 7 to RHEL 8."""

    class FactsCollectionPhase(Phase):
        """
        Get information (facts) about the system (e.g. installed packages, configuration, ...).

        No decision should be done in this phase. Scan the system to get information you need and provide
        it to other actors in the following phases.
        """

        name = 'FactsCollection'
        filter = TagFilter(tags.FactsPhaseTag)
        policies = Policies(Policies.Errors.FailPhase,
                            Policies.Retry.Phase)
        flags = Flags()

    class ChecksPhase(Phase):
        """
        Check upgradability of the system, produce user question if needed and produce output for the report.

        Check whether it is possible to upgrade the system and detect potential risks. It is not expected to get
        additional information about the system in this phase, but rather work with data provided by the actors from
        the FactsCollection. When a potential risk is detected for upgrade, produce messages for the Reports phase.
        """

        name = 'Checks'
        filter = TagFilter(tags.ChecksPhaseTag)
        policies = Policies(Policies.Errors.FailPhase,
                            Policies.Retry.Phase)
        flags = Flags()

    class TargetTransactionFactsCollectionPhase(Phase):
        """
        Get information about target system. Analogy of FactsCollectionPhase for target system.

        Here we can collect information what repositories are available on target system,
        what is expected calculation of target transaction (what will be installed, removed, ...
        """

        name = 'TargetTransactionFactsCollection'
        filter = TagFilter(tags.TargetTransactionFactsPhaseTag)
        policies = Policies(Policies.Errors.FailPhase,
                            Policies.Retry.Phase)
        flags = Flags()

    class TargetTransactionChecksPhase(Phase):
        """
        Checks upgradability regarding the information gathered about the target system.

        Check whether expected repositories and rpms are available, what rpms are planned
        to install, remove, ...

        IOW, checks related to rpm transaction mainly.
        """

        name = 'TargetTransactionCheck'
        filter = TagFilter(tags.TargetTransactionChecksPhaseTag)
        policies = Policies(Policies.Errors.FailPhase,
                            Policies.Retry.Phase)
        flags = Flags()

    class ReportsPhase(Phase):
        """Provide user with the result of the checks."""

        name = 'Reports'
        filter = TagFilter(tags.ReportPhaseTag)
        policies = Policies(Policies.Errors.FailPhase,
                            Policies.Retry.Phase)
        flags = Flags()

    # The following two phases have been removed but are kept here for reference or in case they are needed again

    # class AttachPackageReposPhase(Phase):
    #    name = 'AttachPackageRepos'
    #    #NOTE: in case of use the AttachPackageReposPhaseTag tag has to be created
    #    filter = TagFilter(tags.AttachPackageReposPhaseTag)
    #    policies = Policies(Policies.Errors.FailPhase,
    #                        Policies.Retry.Phase)
    #    flags = Flags()

    # class PlanningPhase(Phase):
    #    name = 'Planning'
    #    #NOTE: in case of use the PlanningPhaseTag tag has to be created
    #    filter = TagFilter(tags.PlanningPhaseTag)
    #    policies = Policies(Policies.Errors.FailPhase,
    #                        Policies.Retry.Phase)
    #    flags = Flags()

    class DownloadPhase(Phase):
        """Download data needed for the upgrade and prepare RPM transaction for the upgrade."""

        name = 'Download'
        filter = TagFilter(tags.DownloadPhaseTag)
        policies = Policies(Policies.Errors.FailPhase,
                            Policies.Retry.Phase)
        flags = Flags()

    class InterimPreparationPhase(Phase):
        """Prepare an initial RAM file system (if required). Setup bootloader."""

        name = 'InterimPreparation'
        filter = TagFilter(tags.InterimPreparationPhaseTag)
        policies = Policies(Policies.Errors.FailImmediately,
                            Policies.Retry.Phase)
        flags = Flags(request_restart_after_phase=True)

    class InitRamStartPhase(Phase):
        """Boot into the upgrade initramfs, mount disks, etc."""

        name = 'InitRamStart'
        filter = TagFilter(tags.InitRamStartPhaseTag)
        policies = Policies(Policies.Errors.FailPhase,
                            Policies.Retry.Phase)
        flags = Flags()

    # The following two phases have been removed but are kept here for reference or in case they are needed again

    # class NetworkPhase(Phase):
    #    name = 'Network'
    #    #NOTE: in case of use the NetworkPhaseTag tag has to be created
    #    filter = TagFilter(tags.NetworkPhaseTag)
    #    policies = Policies(Policies.Errors.FailPhase,
    #                        Policies.Retry.Phase)
    #    flags = Flags()

    # class StoragePhase(Phase):
    #    name = 'Storage'
    #    #NOTE: in case of use the StoragePhaseTag tag has to be created
    #    filter = TagFilter(tags.StoragePhaseTag)
    #    policies = Policies(Policies.Errors.FailPhase,
    #                        Policies.Retry.Phase)
    #    flags = Flags()

    class LateTestsPhase(Phase):
        """Last tests before the RPM upgrade that have to be done with the new kernel and systemd."""

        name = 'LateTests'
        filter = TagFilter(tags.LateTestsPhaseTag)
        policies = Policies(Policies.Errors.FailPhase,
                            Policies.Retry.Phase)
        flags = Flags()

    class PreparationPhase(Phase):
        """Prepare the environment to ascertain success of the RPM upgrade transaction."""

        name = 'Preparation'
        filter = TagFilter(tags.PreparationPhaseTag)
        policies = Policies(Policies.Errors.FailPhase,
                            Policies.Retry.Phase)
        flags = Flags()

    class RPMUpgradePhase(Phase):
        """Perform the RPM transaction, i.e. upgrade the RPMs."""

        name = 'RPMUpgrade'
        filter = TagFilter(tags.RPMUpgradePhaseTag)
        policies = Policies(Policies.Errors.FailPhase,
                            Policies.Retry.Phase)
        flags = Flags(is_checkpoint=True)

    class ApplicationsPhase(Phase):
        """
        Perform the necessary steps to finish upgrade of applications provided by Red Hat.

        This may include moving/renaming of configuration files, modifying configuration of applications to be able
        to run correctly and with as similar behaviour to the original as possible.
        """

        name = 'Applications'
        filter = TagFilter(tags.ApplicationsPhaseTag)
        policies = Policies(Policies.Errors.FailPhase,
                            Policies.Retry.Phase)
        flags = Flags()

    class ThirdPartyApplicationsPhase(Phase):
        """Analogy to the Applications phase, but for third party and custom applications."""

        name = 'ThirdPartyApplications'
        filter = TagFilter(tags.ThirdPartyApplicationsPhaseTag)
        policies = Policies(Policies.Errors.FailPhase,
                            Policies.Retry.Phase)
        flags = Flags()

    class FinalizationPhase(Phase):
        """
        Additional actions that should be done before rebooting into the upgraded system.

        For example SELinux relabeling.
        """

        name = 'Finalization'
        filter = TagFilter(tags.FinalizationPhaseTag)
        policies = Policies(Policies.Errors.FailPhase,
                            Policies.Retry.Phase)
        flags = Flags(restart_after_phase=True)

    class FirstBootPhase(Phase):
        """Actions to be done right after booting into the upgraded system."""

        name = 'FirstBoot'
        filter = TagFilter(tags.FirstBootPhaseTag)
        policies = Policies(Policies.Errors.FailPhase,
                            Policies.Retry.Phase)
        flags = Flags()



import pytest

from leapp.libraries.actor.cupsmigrate import move_directives


class MockFileSystem(object):
    def __init__(self,
                 infiles):
        self.files = {}
        self.ssl_dir = []
        self.files = infiles
        for path in infiles.keys():
            if path.startswith('ssl') or path.startswith('/etc/cups/ssl'):
                self.ssl_dir.append(path.rsplit('/', 1)[1])

    def readlines(self, path):
        if path in self.files.keys():
            return self.files[path].splitlines(True)
        raise IOError('Error when reading file {} - file '
                      'does not exist.'.format(path))

    def write(self, path, mode, content):
        if isinstance(content, list):
            content = ''.join(content)

        if mode == 'w':
            self.files[path] = content
        else:
            self.files[path] += content

    def copy_to_ssl(self, oldpath):
        self.ssl_dir.append(oldpath.rsplit('/', 1)[1])


testdata = (
    (
        {
            '/etc/cups/cupsd.conf': 'ifdfdfgfg\n',
            '/etc/cups/cups-files.conf': 'clean\n'
        },
        {
            '/etc/cups/cupsd.conf': 'ifdfdfgfg\n',
            '/etc/cups/cups-files.conf': 'clean\n'
        }
    ),
    (
        {
            '/etc/cups/cupsd.conf': '#PassEnv smht\nHello world\n',
            '/etc/cups/cups-files.conf': 'clean\n'
        },
        {
            '/etc/cups/cupsd.conf': '#PassEnv smht\nHello world\n',
            '/etc/cups/cups-files.conf': 'clean\n'
        }
    ),
    (
        {
            '/etc/cups/cupsd.conf': 'PassEnv smth\nHello world\n',
            '/etc/cups/cups-files.conf': 'clean\n'
        },
        {
            '/etc/cups/cupsd.conf': 'Hello world\n',
            '/etc/cups/cups-files.conf': 'clean\n\n# added by Leapp\nPassEnv smth\n'
        }
    ),
    (
        {
            '/etc/cups/cupsd.conf': 'SetEnv smht to\nHello world\n',
            '/etc/cups/cups-files.conf': 'clean\n'
        },
        {
            '/etc/cups/cupsd.conf': 'Hello world\n',
            '/etc/cups/cups-files.conf': 'clean\n\n# added by Leapp\nSetEnv smht to\n'
        }
    ),
    (
        {
            '/etc/cups/cupsd.conf': 'PassEnv smht\nSetEnv set to\nHello world\n',
            '/etc/cups/cups-files.conf': 'clean\n'
        },
        {
            '/etc/cups/cupsd.conf': 'Hello world\n',
            '/etc/cups/cups-files.conf': 'clean\n\n# added by Leapp\n'
                                         'PassEnv smht\nSetEnv set to\n'
        }
    ),
    (
        {
            '/etc/cups/cupsd.conf': 'PassEnv smth\nSetEnv set to\nPri'
                                    'ntcapFormat any\nHello world\n',
            '/etc/cups/cups-files.conf': 'clean\n'
        },
        {
            '/etc/cups/cupsd.conf': 'Hello world\n',
            '/etc/cups/cups-files.conf': 'clean\n\n# added by Leapp\n'
                                         'PassEnv smth\nSetEnv set to'
                                         '\nPrintcapFormat any\n'
        }
    )
)


@pytest.mark.parametrize('files,expected', testdata)
def test_move_directives(files, expected):
    op = MockFileSystem(infiles=files)

    move_directives(['PassEnv', 'SetEnv', 'PrintcapFormat'], op)

    assert op.files.get('/etc/cups/cupsd.conf', None) == expected.get('/etc/cups/cupsd.conf', None)
    assert op.files.get('/etc/cups/cups-files.conf', None) == expected.get('/etc/cups/cups-files.conf', None)



from leapp.actors import Actor
from leapp.libraries.actor.scandynamiclinkerconfiguration import scan_dynamic_linker_configuration
from leapp.models import DistributionSignedRPM, DynamicLinkerConfiguration
from leapp.tags import FactsPhaseTag, IPUWorkflowTag


class ScanDynamicLinkerConfiguration(Actor):
    """
    Scan the dynamic linker configuration and find modifications.

    The dynamic linker configuration files can be used to replace standard libraries
    with different custom libraries. The in-place upgrade does not support customization
    of this configuration by user. This actor produces information about detected
    modifications.
    """

    name = 'scan_dynamic_linker_configuration'
    consumes = (DistributionSignedRPM,)
    produces = (DynamicLinkerConfiguration,)
    tags = (FactsPhaseTag, IPUWorkflowTag)

    def process(self):
        scan_dynamic_linker_configuration()



from leapp import reporting
from leapp.actors import Actor
from leapp.libraries.common.rpms import has_package
from leapp.models import DistributionSignedRPM
from leapp.reporting import create_report, Report
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class CheckIrssi(Actor):
    """
    Check if irssi is installed. If yes, write information about non-compatible changes.
    """

    name = 'checkirssi'
    consumes = (DistributionSignedRPM,)
    produces = (Report,)
    tags = (ChecksPhaseTag, IPUWorkflowTag)

    def process(self):
        if has_package(DistributionSignedRPM, 'irssi'):
            create_report([
                reporting.Title('Irssi incompatible changes in the next major version'),
                reporting.Summary(
                    'Disabled support for the insecure SSLv2 protocol.\n'
                    'Disabled SSLv3 due to the POODLE vulnerability.\n'
                    'Removing networks will now remove all attached servers and channels.\n'
                    'Removed --disable-ipv6 option.\n'
                ),
                reporting.Severity(reporting.Severity.LOW),
                reporting.Groups([
                        reporting.Groups.COMMUNICATION,
                        reporting.Groups.TOOLS
                ]),
                reporting.Remediation(hint='Please update your scripts to be compatible with the changes.'),
                reporting.RelatedResource('package', 'irssi')
            ])



import pytest

from leapp.libraries import stdlib
from leapp.libraries.actor import setuptargetrepos
from leapp.libraries.common.testutils import CurrentActorMocked, produce_mocked
from leapp.libraries.stdlib import api
from leapp.models import (
    CustomTargetRepository,
    InstalledRPM,
    PESIDRepositoryEntry,
    RepoMapEntry,
    RepositoriesBlacklisted,
    RepositoriesFacts,
    RepositoriesMapping,
    RepositoriesSetupTasks,
    RepositoryData,
    RepositoryFile,
    RPM,
    TargetRepositories
)

RH_PACKAGER = 'Red Hat, Inc. <http://bugzilla.redhat.com/bugzilla>'


def mock_package(pkg_name, repository=None):
    return RPM(name=pkg_name, version='0.1', release='1.sm01', epoch='1', packager=RH_PACKAGER, arch='noarch',
               pgpsig='RSA/SHA256, Mon 01 Jan 1970 00:00:00 AM -03, Key ID 199e2f91fd431d51', repository=repository)


def test_minimal_execution(monkeypatch):
    """
    Tests whether the actor does not fail if no messages except the RepositoriesMapping are provided.
    """
    msgs = [
        RepositoriesMapping(mapping=[], repositories=[])
    ]

    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(msgs=msgs))
    monkeypatch.setattr(api, 'produce', produce_mocked())

    setuptargetrepos.process()


def test_custom_repos(monkeypatch):
    """
    Tests whether the CustomRepos provided to the actor are propagated to the TargetRepositories after
    blacklist filtering is applied on them.
    """
    custom = CustomTargetRepository(repoid='rhel-8-server-rpms',
                                    name='RHEL 8 Server (RPMs)',
                                    baseurl='https://.../dist/rhel/server/8/os',
                                    enabled=True)

    blacklisted = CustomTargetRepository(repoid='rhel-8-blacklisted-rpms',
                                         name='RHEL 8 Blacklisted (RPMs)',
                                         baseurl='https://.../dist/rhel/blacklisted/8/os',
                                         enabled=True)

    repos_blacklisted = RepositoriesBlacklisted(repoids=['rhel-8-blacklisted-rpms'])

    repositories_mapping = RepositoriesMapping(
        mapping=[],
        repositories=[]
    )

    msgs = [custom, blacklisted, repos_blacklisted, repositories_mapping]

    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(msgs=msgs))
    monkeypatch.setattr(api, 'produce', produce_mocked())

    setuptargetrepos.process()

    assert api.produce.called

    custom_repos = api.produce.model_instances[0].custom_repos
    assert len(custom_repos) == 1
    assert custom_repos[0].repoid == 'rhel-8-server-rpms'


def test_repositories_setup_tasks(monkeypatch):
    """
    Tests whether the actor propagates repositories received via a RepositoriesSetupTasks message
    to the resulting TargetRepositories (and blacklist filtering is applied to them).
    """
    repositories_setup_tasks = RepositoriesSetupTasks(to_enable=['rhel-8-server-rpms',
                                                                 'rhel-8-blacklisted-rpms'])
    repos_blacklisted = RepositoriesBlacklisted(repoids=['rhel-8-blacklisted-rpms'])
    repositories_mapping = RepositoriesMapping(mapping=[], repositories=[])
    msgs = [repositories_setup_tasks, repos_blacklisted, repositories_mapping]

    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(msgs=msgs))
    monkeypatch.setattr(api, 'produce', produce_mocked())

    setuptargetrepos.process()

    assert api.produce.called

    rhel_repos = api.produce.model_instances[0].rhel_repos
    assert len(rhel_repos) == 1
    assert rhel_repos[0].repoid == 'rhel-8-server-rpms'


def test_repos_mapping(monkeypatch):
    """
    Tests whether actor correctly determines what repositories should be enabled on target based
    on the information about what repositories are enabled on the source system using
    the RepositoriesMapping information.
    """
    repos_data = [
        RepositoryData(repoid='rhel-7-server-rpms', name='RHEL 7 Server'),
        RepositoryData(repoid='rhel-7-blacklisted-rpms', name='RHEL 7 Blacklisted')]

    repos_files = [RepositoryFile(file='/etc/yum.repos.d/redhat.repo', data=repos_data)]
    facts = RepositoriesFacts(repositories=repos_files)
    installed_rpms = InstalledRPM(
        items=[mock_package('foreman', 'rhel-7-for-x86_64-satellite-extras-rpms'),
               mock_package('foreman-proxy', 'nosuch-rhel-7-for-x86_64-satellite-extras-rpms')])

    repomap = RepositoriesMapping(
        mapping=[RepoMapEntry(source='rhel7-base', target=['rhel8-baseos', 'rhel8-appstream', 'rhel8-blacklist']),
                 RepoMapEntry(source='rhel7-satellite-extras', target=['rhel8-satellite-extras'])],
        repositories=[
            PESIDRepositoryEntry(
                pesid='rhel7-base',
                repoid='rhel-7-server-rpms',
                major_version='7',
                arch='x86_64',
                repo_type='rpm',
                channel='ga',
                rhui=''
            ),
            PESIDRepositoryEntry(
                pesid='rhel8-baseos',
                repoid='rhel-8-for-x86_64-baseos-htb-rpms',
                major_version='8',
                arch='x86_64',
                repo_type='rpm',
                channel='ga',
                rhui=''
            ),
            PESIDRepositoryEntry(
                pesid='rhel8-appstream',
                repoid='rhel-8-for-x86_64-appstream-htb-rpms',
                major_version='8',
                arch='x86_64',
                repo_type='rpm',
                channel='ga',
                rhui=''
            ),
            PESIDRepositoryEntry(
                pesid='rhel8-blacklist',
                repoid='rhel-8-blacklisted-rpms',
                major_version='8',
                arch='x86_64',
                repo_type='rpm',
                channel='ga',
                rhui=''
            ),
            PESIDRepositoryEntry(
                pesid='rhel7-satellite-extras',
                repoid='rhel-7-for-x86_64-satellite-extras-rpms',
                major_version='7',
                arch='x86_64',
                repo_type='rpm',
                channel='ga',
                rhui=''
            ),
            PESIDRepositoryEntry(
                pesid='rhel8-satellite-extras',
                repoid='rhel-8-for-x86_64-satellite-extras-rpms',
                major_version='8',
                arch='x86_64',
                repo_type='rpm',
                channel='ga',
                rhui=''
            ),
        ]
    )

    repos_blacklisted = RepositoriesBlacklisted(repoids=['rhel-8-blacklisted-rpms'])

    msgs = [facts, repomap, repos_blacklisted, installed_rpms]

    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(msgs=msgs))
    monkeypatch.setattr(api, 'produce', produce_mocked())

    setuptargetrepos.process()
    assert api.produce.called

    rhel_repos = api.produce.model_instances[0].rhel_repos
    assert len(rhel_repos) == 3

    produced_rhel_repoids = {repo.repoid for repo in rhel_repos}
    expected_rhel_repoids = {'rhel-8-for-x86_64-baseos-htb-rpms', 'rhel-8-for-x86_64-appstream-htb-rpms',
                             'rhel-8-for-x86_64-satellite-extras-rpms'}
    assert produced_rhel_repoids == expected_rhel_repoids
